"""
"""
import os
import re
from typing import Mapping, Callable, Tuple, List

import pkg_resources
from AnyQt.QtCore import Qt
from AnyQt.QtGui import QPalette, QColor


def _make_palette(
        base: QColor, text: QColor, window: QColor,
        highlight: QColor,
        highlight_disabled: QColor,
        text_disabled: QColor, link: QColor,
        light: QColor, mid: QColor, dark: QColor, shadow: QColor

):
    palette = QPalette()
    palette.setColor(QPalette.Window, window)
    palette.setColor(QPalette.WindowText, text)
    palette.setColor(QPalette.Disabled, QPalette.WindowText, text_disabled)
    palette.setColor(QPalette.Base, base)
    palette.setColor(QPalette.AlternateBase, window)
    palette.setColor(QPalette.ToolTipBase, window)
    palette.setColor(QPalette.ToolTipText, text)
    palette.setColor(QPalette.Text, text)
    palette.setColor(QPalette.Disabled, QPalette.Text, text_disabled)
    palette.setColor(QPalette.Button, window)
    palette.setColor(QPalette.ButtonText, text)
    palette.setColor(QPalette.Disabled, QPalette.ButtonText, text_disabled)
    palette.setColor(QPalette.BrightText, Qt.white)
    palette.setColor(QPalette.Highlight, highlight)
    palette.setColor(QPalette.Disabled, QPalette.Highlight, highlight_disabled)
    palette.setColor(QPalette.HighlightedText, text)
    palette.setColor(QPalette.Light, light)
    palette.setColor(QPalette.Mid, mid)
    palette.setColor(QPalette.Dark, dark)
    palette.setColor(QPalette.Shadow, shadow)
    palette.setColor(QPalette.Link, link)
    return palette


def breeze_light() -> QPalette:
    # 'Breeze-Light' color scheme from KDE.
    return _make_palette(
        text=QColor("#30363C"),
        text_disabled=QColor("#888786"),
        window=QColor("#EFF0F1"),
        base=QColor("#FCFCFC"),
        highlight=QColor("#00B0EF"),
        highlight_disabled=QColor("#00B0EF"),
        link=QColor("#0057B4"),
        light=QColor("#ffffff"),
        mid=QColor("#c4c9cd"),
        dark=QColor("#888e93"),
        shadow=QColor("#474a4c"),
    )


def breeze_dark() -> QPalette:
    # 'Breeze Dark' color scheme from KDE.
    return _make_palette(
        text=QColor(239, 240, 241),
        text_disabled=QColor(98, 108, 118),
        window=QColor(49, 54, 59),
        base=QColor(35, 38, 41),
        highlight=QColor(61, 174, 233),
        highlight_disabled=QColor(61, 174, 233),
        link=QColor(41, 128, 185),
        light=QColor(69, 76, 84),
        mid=QColor(43, 47, 52),
        dark=QColor(28, 31, 34),
        shadow=QColor(20, 23, 25),
    )


def zion_reversed() -> QPalette:
    # 'Zion Reversed' color scheme from KDE.
    window = QColor(16, 16, 16)
    return _make_palette(
        text=QColor(Qt.white),
        text_disabled=QColor(85, 85, 85),
        window=window,
        base=QColor(Qt.black),
        highlight=QColor(0, 49, 110),
        highlight_disabled=window,
        link=QColor(128, 181, 255),
        light=QColor(174, 174, 174),
        mid=QColor(89, 89, 89),
        dark=QColor(118, 118, 118),
        shadow=QColor(141, 141, 141),
    )


def dark():
    window = QColor(0x30, 0x30, 0x30)
    return _make_palette(
        text=QColor(Qt.white),
        base=QColor(0x20, 0x20, 0x20),
        window=QColor(0x30, 0x30, 0x30),
        text_disabled=QColor(0x9B, 0x9B, 0x9B),
        highlight=QColor(0x2E, 0x93, 0xFF),
        highlight_disabled=window,
        link=QColor(0x2E, 0x93, 0xFF),
        light=QColor(174, 174, 174),
        mid=QColor(89, 89, 89),
        dark=QColor(118, 118, 118),
        shadow=QColor(141, 141, 141),
    )


colorthemes = {
    "breeze-light": breeze_light,
    "breeze-dark": breeze_dark,
    "zion-reversed": zion_reversed,
    "dark": dark
}  # type: Mapping[str, Callable[[],QPalette]]


def style_sheet(stylesheet: str) -> Tuple[str, List[Tuple[str, str]]]:
    """
    Load and return a stylesheet string from path.

    Extract special `@prefix: subdirname` 'directives' and return the
    (prefix, dirname) tuples. These should be added to `QDir.searchPath` in
    order to locate resources.

    Parameters
    ----------
    stylesheet: str
        A path to a css (Qt's stylesheet) file. Can be a relative path w.r.t.
        this package's directory.

    Returns
    -------
    stylesheet: str
    searchpaths: List[Tuple[str, str]]
    """
    def process_qss(content: str, base: str):
        pattern = re.compile(
            r"^\s*@([a-zA-Z0-9_]+?)\s*:\s*([a-zA-Z0-9_/]+?);\s*$",
            flags=re.MULTILINE
        )
        matches = pattern.findall(content)
        paths = []
        for prefix, search_path in matches:
            paths.append((prefix, os.path.join(base, search_path)))
        content = pattern.sub("", content)
        return content, paths

    stylesheet_string = None
    try:
        with open(stylesheet, "r", encoding="utf-8") as f:
            stylesheet_string = f.read()
    except (OSError, UnicodeDecodeError):
        pass
    else:
        return process_qss(stylesheet_string, os.path.basename(stylesheet))

    if not os.path.splitext(stylesheet)[1]:
        # no extension
        stylesheet = os.path.extsep.join([stylesheet, "qss"])

    pkg_name = __package__
    resource = stylesheet

    if pkg_resources.resource_exists(pkg_name, resource):
        stylesheet_string = \
            pkg_resources.resource_string(pkg_name, resource).decode("utf-8")
        base = pkg_resources.resource_filename(pkg_name, "")
        return process_qss(stylesheet_string, base)
    return stylesheet_string, []



"""
Widget Tool Box
===============


A tool box with a tool grid for each category.

"""
from typing import Optional, Iterable, Any

from AnyQt.QtWidgets import (
    QAbstractButton, QSizePolicy, QAction, QApplication, QToolButton,
    QWidget
)
from AnyQt.QtGui import (
    QDrag, QPalette, QBrush, QIcon, QColor, QGradient, QActionEvent,
    QMouseEvent
)
from AnyQt.QtCore import (
    Qt, QObject, QAbstractItemModel, QModelIndex, QSize, QEvent, QMimeData,
    QByteArray, QDataStream, QIODevice, QPoint
)
from AnyQt.QtCore import pyqtSignal as Signal, pyqtProperty as Property

from ..gui.toolbox import ToolBox
from ..gui.toolgrid import ToolGrid
from ..gui.quickhelp import StatusTipPromoter
from ..gui.utils import create_gradient
from ..registry.qt import QtWidgetRegistry


def iter_index(model, index):
    # type: (QAbstractItemModel, QModelIndex) -> Iterable[QModelIndex]
    """
    Iterate over child indexes of a `QModelIndex` in a `model`.
    """
    for row in range(model.rowCount(index)):
        yield model.index(row, 0, index)


def item_text(index):  # type: (QModelIndex) -> str
    value = index.data(Qt.DisplayRole)
    if value is None:
        return ""
    else:
        return str(value)


def item_icon(index):  # type: (QModelIndex) -> QIcon
    value = index.data(Qt.DecorationRole)
    if isinstance(value, QIcon):
        return value
    else:
        return QIcon()


def item_tooltip(index):  # type: (QModelIndex) -> str
    value = index.data(Qt.ToolTipRole)
    if isinstance(value, str):
        return value
    return item_text(index)


def item_background(index):  # type: (QModelIndex) -> Optional[QBrush]
    value = index.data(Qt.BackgroundRole)
    if isinstance(value, QBrush):
        return value
    elif isinstance(value, (QColor, Qt.GlobalColor, QGradient)):
        return QBrush(value)
    else:
        return None


class WidgetToolGrid(ToolGrid):
    """
    A Tool Grid with widget buttons. Populates the widget buttons
    from a item model. Also adds support for drag operations.

    """
    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None
        super().__init__(*args, **kwargs)

        self.__model = None               # type: Optional[QAbstractItemModel]
        self.__rootIndex = QModelIndex()  # type: QModelIndex
        self.__actionRole = QtWidgetRegistry.WIDGET_ACTION_ROLE  # type: int

        self.__dragListener = DragStartEventListener(self)
        self.__dragListener.dragStartOperationRequested.connect(
            self.__startDrag
        )
        self.__statusTipPromoter = StatusTipPromoter(self)

    def setModel(self, model, rootIndex=QModelIndex()):
        # type: (QAbstractItemModel, QModelIndex) -> None
        """
        Set a model (`QStandardItemModel`) for the tool grid. The
        widget actions are children of the rootIndex.

        .. warning:: The model should not be deleted before the
                     `WidgetToolGrid` instance.

        """
        if self.__model is not None:
            self.__model.rowsInserted.disconnect(self.__on_rowsInserted)
            self.__model.rowsRemoved.disconnect(self.__on_rowsRemoved)
            self.__model = None

        self.__model = model
        self.__rootIndex = rootIndex

        if self.__model is not None:
            self.__model.rowsInserted.connect(self.__on_rowsInserted)
            self.__model.rowsRemoved.connect(self.__on_rowsRemoved)

        self.__initFromModel(model, rootIndex)

    def model(self):  # type: () -> Optional[QAbstractItemModel]
        """
        Return the model for the tool grid.
        """
        return self.__model

    def rootIndex(self):  # type: () -> QModelIndex
        """
        Return the root index of the model.
        """
        return self.__rootIndex

    def setActionRole(self, role):
        # type: (int) -> None
        """
        Set the action role. This is the model role containing a
        `QAction` instance.
        """
        if self.__actionRole != role:
            self.__actionRole = role
            if self.__model:
                self.__update()

    def actionRole(self):  # type: () -> int
        """
        Return the action role.
        """
        return self.__actionRole

    def actionEvent(self, event):  # type: (QActionEvent) -> None
        if event.type() == QEvent.ActionAdded:
            # Creates and inserts the button instance.
            super().actionEvent(event)

            button = self.buttonForAction(event.action())
            button.installEventFilter(self.__dragListener)
            button.installEventFilter(self.__statusTipPromoter)
            return
        elif event.type() == QEvent.ActionRemoved:
            button = self.buttonForAction(event.action())
            button.removeEventFilter(self.__dragListener)
            button.removeEventFilter(self.__statusTipPromoter)

            # Removes the button
            super().actionEvent(event)
            return
        else:
            super().actionEvent(event)

    def __initFromModel(self, model, rootIndex):
        # type: (QAbstractItemModel, QModelIndex) -> None
        """
        Initialize the grid from the model with rootIndex as the root.
        """
        for i, index in enumerate(iter_index(model, rootIndex)):
            self.__insertItem(i, index)

    def __insertItem(self, index, item):
        # type: (int, QModelIndex) -> None
        """
        Insert a widget action from `item` (`QModelIndex`) at `index`.
        """
        value = item.data(self.__actionRole)
        if isinstance(value, QAction):
            action = value
        else:
            action = QAction(item_text(item), self)
            action.setIcon(item_icon(item))
            action.setToolTip(item_tooltip(item))

        self.insertAction(index, action)

    def __update(self):  # type: () -> None
        self.clear()
        if self.__model is not None:
            self.__initFromModel(self.__model, self.__rootIndex)

    def __on_rowsInserted(self, parent, start, end):
        # type: (QModelIndex, int, int) -> None
        """
        Insert items from range start:end into the grid.
        """
        if parent == self.__rootIndex:
            for i in range(start, end + 1):
                item = self.__model.index(i, 0, self.__rootIndex)
                self.__insertItem(i, item)

    def __on_rowsRemoved(self, parent, start, end):
        # type: (QModelIndex, int, int) -> None
        """
        Remove items from range start:end from the grid.
        """
        if parent == self.__rootIndex:
            for i in reversed(range(start - 1, end)):
                action = self.actions()[i]
                self.removeAction(action)

    def __startDrag(self, button):
        # type: (QToolButton) -> None
        """
        Start a drag from button
        """
        action = button.defaultAction()
        desc = action.data()  # Widget Description
        icon = action.icon()
        drag_data = QMimeData()
        drag_data.setData(
            "application/vnd.orange-canvas.registry.qualified-name",
            desc.qualified_name.encode("utf-8")
        )
        drag = QDrag(button)
        drag.setPixmap(icon.pixmap(self.iconSize()))
        drag.setMimeData(drag_data)
        drag.exec(Qt.CopyAction)


class DragStartEventListener(QObject):
    """
    An event filter object that can be used to detect drag start
    operation on buttons which otherwise do not support it.

    """
    dragStartOperationRequested = Signal(QAbstractButton)
    """A drag operation started on a button."""

    def __init__(self, parent=None, **kwargs):
        # type: (Optional[QObject], Any) -> None
        super().__init__(parent, **kwargs)
        self.button = None         # type: Optional[Qt.MouseButton]
        self.buttonDownObj = None  # type: Optional[QAbstractButton]
        self.buttonDownPos = None  # type: Optional[QPoint]

    def eventFilter(self, obj, event):
        # type: (QObject, QEvent) -> bool
        if event.type() == QEvent.MouseButtonPress:
            assert isinstance(event, QMouseEvent)
            self.buttonDownPos = event.pos()
            self.buttonDownObj = obj
            self.button = event.button()

        elif event.type() == QEvent.MouseMove and obj is self.buttonDownObj:
            assert self.buttonDownObj is not None
            if (self.buttonDownPos - event.pos()).manhattanLength() > \
                    QApplication.startDragDistance() and \
                    not self.buttonDownObj.hitButton(event.pos()):
                # Process the widget's mouse event, before starting the
                # drag operation, so the widget can update its state.
                obj.mouseMoveEvent(event)
                self.dragStartOperationRequested.emit(obj)

                obj.setDown(False)

                self.button = None
                self.buttonDownPos = None
                self.buttonDownObj = None
                return True  # Already handled

        return super().eventFilter(obj, event)


class WidgetToolBox(ToolBox):
    """
    `WidgetToolBox` widget shows a tool box containing button grids of
    actions for a :class:`QtWidgetRegistry` item model.
    """

    triggered = Signal(QAction)
    hovered = Signal(QAction)

    def __init__(self, parent=None):
        # type: (Optional[QWidget]) -> None
        super().__init__(parent)
        self.__model = None  # type: Optional[QAbstractItemModel]
        self.__iconSize = QSize(25, 25)
        self.__buttonSize = QSize(50, 50)
        self.setSizePolicy(QSizePolicy.Fixed,
                           QSizePolicy.Expanding)

    def setIconSize(self, size):  # type: (QSize) -> None
        """
        Set the widget icon size (icons in the button grid).
        """
        if self.__iconSize != size:
            self.__iconSize = QSize(size)
            for widget in map(self.widget, range(self.count())):
                widget.setIconSize(size)

    def iconSize(self):  # type: () -> QSize
        """
        Return the widget buttons icon size.
        """
        return QSize(self.__iconSize)

    iconSize_ = Property(QSize, fget=iconSize, fset=setIconSize,
                         designable=True)

    def setButtonSize(self, size):  # type: (QSize) -> None
        """
        Set fixed widget button size.
        """
        if self.__buttonSize != size:
            self.__buttonSize = QSize(size)
            for widget in map(self.widget, range(self.count())):
                widget.setButtonSize(size)

    def buttonSize(self):  # type: () -> QSize
        """Return the widget button size
        """
        return QSize(self.__buttonSize)

    buttonSize_ = Property(QSize, fget=buttonSize, fset=setButtonSize,
                           designable=True)

    def saveState(self):  # type: () -> QByteArray
        """
        Return the toolbox state (as a `QByteArray`).

        .. note:: Individual tabs are stored by their action's text.

        """
        version = 2

        actions = map(self.tabAction, range(self.count()))
        expanded = [action for action in actions if action.isChecked()]
        expanded = [action.text() for action in expanded]

        byte_array = QByteArray()
        stream = QDataStream(byte_array, QIODevice.WriteOnly)
        stream.writeInt(version)
        stream.writeQStringList(expanded)

        return byte_array

    def restoreState(self, state):  # type: (QByteArray) -> bool
        """
        Restore the toolbox from a :class:`QByteArray` `state`.

        .. note:: The toolbox should already be populated for the state
                  changes to take effect.

        """
        stream = QDataStream(state, QIODevice.ReadOnly)
        version = stream.readInt()
        if version == 2:
            expanded = stream.readQStringList()
            for action in map(self.tabAction, range(self.count())):
                if (action.text() in expanded) != action.isChecked():
                    action.trigger()
            return True
        return False

    def setModel(self, model):
        # type: (QAbstractItemModel) -> None
        """
        Set the widget registry model (:class:`QAbstractItemModel`) for
        this toolbox.
        """
        if self.__model is not None:
            self.__model.dataChanged.disconnect(self.__on_dataChanged)
            self.__model.rowsInserted.disconnect(self.__on_rowsInserted)
            self.__model.rowsRemoved.disconnect(self.__on_rowsRemoved)

        self.__model = model
        if self.__model is not None:
            self.__model.dataChanged.connect(self.__on_dataChanged)
            self.__model.rowsInserted.connect(self.__on_rowsInserted)
            self.__model.rowsRemoved.connect(self.__on_rowsRemoved)

        self.__initFromModel(self.__model)

    def __initFromModel(self, model):
        # type: (QAbstractItemModel) -> None
        for row in range(model.rowCount()):
            self.__insertItem(model.index(row, 0), self.count())

    def __insertItem(self, item, index):
        # type: (QModelIndex, int) -> None
        """
        Insert category item  (`QModelIndex`) at index.
        """
        grid = WidgetToolGrid()
        grid.setModel(item.model(), item)
        grid.actionTriggered.connect(self.triggered)
        grid.actionHovered.connect(self.hovered)

        grid.setIconSize(self.__iconSize)
        grid.setButtonSize(self.__buttonSize)
        grid.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        text = item_text(item)
        icon = item_icon(item)
        tooltip = item_tooltip(item)

        # Set the 'tab-title' property to text.
        grid.setProperty("tab-title", text)
        grid.setObjectName("widgets-toolbox-grid")

        self.insertItem(index, grid, text, icon, tooltip)
        button = self.tabButton(index)

        # Set the 'highlight' color if applicable
        highlight_foreground = None
        highlight = item_background(item)
        if highlight is None \
                and item.data(QtWidgetRegistry.BACKGROUND_ROLE) is not None:
            highlight = item.data(QtWidgetRegistry.BACKGROUND_ROLE)

        if isinstance(highlight, QBrush) and highlight.style() != Qt.NoBrush:
            if not highlight.gradient():
                value = highlight.color().value()
                gradient = create_gradient(highlight.color())
                highlight = QBrush(gradient)
                highlight_foreground = Qt.black if value > 128 else Qt.white

        palette = button.palette()

        if highlight is not None:
            palette.setBrush(QPalette.Highlight, highlight)
        if highlight_foreground is not None:
            palette.setBrush(QPalette.HighlightedText, highlight_foreground)
        button.setPalette(palette)

    def __on_dataChanged(self, topLeft, bottomRight):
        # type: (QModelIndex, QModelIndex) -> None
        parent = topLeft.parent()
        if not parent.isValid():
            for row in range(topLeft.row(), bottomRight.row() + 1):
                item = topLeft.sibling(row, topLeft.column())
                button = self.tabButton(row)
                button.setIcon(item_icon(item))
                button.setText(item_text(item))
                button.setToolTip(item_tooltip(item))

    def __on_rowsInserted(self, parent, start, end):
        # type: (QModelIndex, int, int) -> None
        """
        Items have been inserted in the model.
        """
        # Only the top level items (categories) are handled here.
        assert self.__model is not None
        if not parent.isValid():
            for i in range(start, end + 1):
                item = self.__model.index(i, 0)
                self.__insertItem(item, i)

    def __on_rowsRemoved(self, parent, start, end):
        # type: (QModelIndex, int, int) -> None
        """
        Rows have been removed from the model.
        """
        # Only the top level items (categories) are handled here.
        if not parent.isValid():
            for i in range(end, start - 1, -1):
                self.removeItem(i)



"""
Example workflows discovery.
"""
import os
import logging
import types

from typing import List, Optional, IO

import pkg_resources

from orangecanvas import config as _config

log = logging.getLogger(__name__)


def list_workflows(package):
    # type: (types.ModuleType) -> List[str]
    """
    Return a list of .ows files in the located next to `package`.
    """

    def is_ows(filename):
        # type: (str) -> bool
        return filename.endswith(".ows")

    resources = pkg_resources.resource_listdir(package.__name__, ".")
    return sorted(filter(is_ows, resources))


def workflows(config=None):
    # type: (Optional[_config.Config]) -> List[ExampleWorkflow]
    """
    Return all known example workflows.
    """
    if config is None:
        config = _config.default

    workflows = []  # type: List[ExampleWorkflow]
    if hasattr(config, "tutorials_entry_points") and \
            callable(config.tutorials_entry_points):
        # back compatibility
        examples_entry_points = config.tutorials_entry_points
    else:
        examples_entry_points = config.examples_entry_points
    for ep in examples_entry_points():
        try:
            examples = ep.resolve()
        except pkg_resources.DistributionNotFound as ex:
            log.warning("Could not load examples from %r (%r)",
                        ep.dist, ex)
            continue
        except Exception:
            log.error("Could not load examples from %r",
                      ep.dist, exc_info=True)
            continue

        if isinstance(examples, types.ModuleType):
            package = examples
            examples = [ExampleWorkflow(t, package, ep.dist)
                        for t in list_workflows(package)]
        elif isinstance(examples, (types.FunctionType, types.MethodType)):
            try:
                examples = examples()
            except Exception as ex:
                log.error("A callable entry point (%r) raised an "
                          "unexpected error.",
                          ex, exc_info=True)
                continue
            examples = [ExampleWorkflow(t, package=None, distribution=ep.dist)
                        for t in examples]
        workflows.extend(examples)
    return workflows


class ExampleWorkflow:
    def __init__(self, resource, package=None, distribution=None):
        # type: (str, Optional[types.ModuleType], Optional[pkg_resources.Distribution]) -> None
        self.resource = resource
        self.package = package
        self.distribution = distribution

    def abspath(self):
        # type: () -> str
        """
        Return absolute filename for the workflow if possible else
        raise an ValueError.
        """
        if self.package is not None:
            return pkg_resources.resource_filename(self.package.__name__,
                                                   self.resource)
        elif isinstance(self.resource, str):
            if os.path.isabs(self.resource):
                return self.resource

        raise ValueError("cannot resolve resource to an absolute name")

    def stream(self):
        # type: () -> IO[bytes]
        """
        Return the example file as an open stream.
        """
        if self.package is not None:
            return pkg_resources.resource_stream(self.package.__name__,
                                                 self.resource)
        elif isinstance(self.resource, str):
            if os.path.isabs(self.resource) and os.path.exists(self.resource):
                return open(self.resource, "rb")

        raise ValueError



"""
===============
Tool Box Widget
===============

A reimplementation of the :class:`QToolBox` widget that keeps all the tabs
in a single :class:`QScrollArea` instance and can keep multiple open tabs.
"""
import enum
from operator import eq, attrgetter

import typing
from typing import NamedTuple, List, Iterable, Optional, Any, Callable

from AnyQt.QtWidgets import (
    QWidget, QFrame, QSizePolicy, QStyle, QStyleOptionToolButton,
    QScrollArea, QVBoxLayout, QToolButton,
    QAction, QActionGroup, QApplication, QAbstractButton, QWIDGETSIZE_MAX,
)
from AnyQt.QtGui import (
    QIcon, QFontMetrics, QPainter, QPalette, QBrush, QPen, QColor, QFont
)
from AnyQt.QtCore import (
    Qt, QObject, QSize, QRect, QPoint, QSignalMapper
)
from AnyQt.QtCore import Signal, Property

from ..utils import set_flag
from .utils import brush_darker, ScrollBar

__all__ = [
    "ToolBox"
]

_ToolBoxPage = NamedTuple(
    "_ToolBoxPage", [
        ("index", int),
        ("widget", QWidget),
        ("action", QAction),
        ("button", QAbstractButton),
    ]
)


class ToolBoxTabButton(QToolButton):
    """
    A tab button for an item in a :class:`ToolBox`.
    """
    class TabPosition(enum.IntFlag):
        Beginning = 0
        Middle = 1
        End = 2
        OnlyOneTab = 3

    Beginning = TabPosition.Beginning
    Middle = TabPosition.Middle
    End = TabPosition.End
    OnlyOneTab = TabPosition.OnlyOneTab

    class SelectedPosition(enum.IntFlag):
        NotAdjacent = 0
        NextIsSelected = 1
        PreviousIsSelected = 2

    NotAdjacent = SelectedPosition.NotAdjacent
    NextIsSelected = SelectedPosition.NextIsSelected
    PreviousIsSelected = SelectedPosition.PreviousIsSelected

    def setNativeStyling(self, state):
        # type: (bool) -> None
        """
        Render tab buttons as native (or css styled) :class:`QToolButtons`.
        If set to `False` (default) the button is pained using a custom
        paint routine.
        """
        self.__nativeStyling = state
        self.update()

    def nativeStyling(self):
        # type: () -> bool
        """
        Use :class:`QStyle`'s to paint the class:`QToolButton` look.
        """
        return self.__nativeStyling

    nativeStyling_ = Property(bool,
                              fget=nativeStyling,
                              fset=setNativeStyling,
                              designable=True)

    def __init__(self, parent=None, **kwargs):
        # type: (Optional[QWidget], Any) -> None
        self.__nativeStyling = False
        self.position = ToolBoxTabButton.OnlyOneTab
        self.selected = ToolBoxTabButton.NotAdjacent
        font = kwargs.pop("font", None)  # type: Optional[QFont]
        super().__init__(parent, **kwargs)

        if font is None:
            self.setFont(QApplication.font("QAbstractButton"))
            self.setAttribute(Qt.WA_SetFont, False)
        else:
            self.setFont(font)

    def enterEvent(self, event):
        super().enterEvent(event)
        self.update()

    def leaveEvent(self, event):
        super().leaveEvent(event)
        self.update()

    def paintEvent(self, event):
        if self.__nativeStyling:
            super().paintEvent(event)
        else:
            self.__paintEventNoStyle()

    def __paintEventNoStyle(self):
        p = QPainter(self)
        opt = QStyleOptionToolButton()
        self.initStyleOption(opt)

        fm = QFontMetrics(opt.font)
        palette = opt.palette

        # highlight brush is used as the background for the icon and background
        # when the tab is expanded and as mouse hover color (lighter).
        brush_highlight = palette.highlight()
        foregroundrole = QPalette.ButtonText
        if opt.state & QStyle.State_Sunken:
            # State 'down' pressed during a mouse press (slightly darker).
            background_brush = brush_darker(brush_highlight, 110)
            foregroundrole = QPalette.HighlightedText
        elif opt.state & QStyle.State_MouseOver:
            background_brush = brush_darker(brush_highlight, 95)
            foregroundrole = QPalette.HighlightedText
        elif opt.state & QStyle.State_On:
            background_brush = brush_highlight
            foregroundrole = QPalette.HighlightedText
        else:
            # The default button brush.
            background_brush = palette.button()

        rect = opt.rect

        icon_area_rect = QRect(rect)
        icon_area_rect.setWidth(int(icon_area_rect.height() * 1.26))

        text_rect = QRect(rect)
        text_rect.setLeft(icon_area_rect.x() + icon_area_rect.width() + 10)

        # Background

        # TODO: Should the tab button have native toolbutton shape, drawn
        #       using PE_PanelButtonTool or even QToolBox tab shape
        # Default outline pen
        pen = QPen(palette.color(QPalette.Mid))

        p.save()
        p.setPen(Qt.NoPen)
        p.setBrush(QBrush(background_brush))
        p.drawRect(rect)

        # Draw the background behind the icon if the background_brush
        # is different.
        if not opt.state & QStyle.State_On:
            p.setBrush(brush_highlight)
            p.drawRect(icon_area_rect)
            # Line between the icon and text
            p.setPen(pen)
            p.drawLine(
                icon_area_rect.x() + icon_area_rect.width(), icon_area_rect.y(),
                icon_area_rect.x() + icon_area_rect.width(),
                icon_area_rect.y() + icon_area_rect.height())

        if opt.state & QStyle.State_HasFocus:
            # Set the focus frame pen and draw the border
            pen = QPen(QColor(brush_highlight))
            p.setPen(pen)
            p.setBrush(Qt.NoBrush)
            # Adjust for pen
            rect = rect.adjusted(0, 0, -1, -1)
            p.drawRect(rect)

        else:
            p.setPen(pen)
            # Draw the top/bottom border
            if self.position == ToolBoxTabButton.OnlyOneTab or \
                    self.position == ToolBoxTabButton.Beginning or \
                    self.selected & ToolBoxTabButton.PreviousIsSelected:
                p.drawLine(rect.x(), rect.y(),
                           rect.x() + rect.width(), rect.y())
            p.drawLine(rect.x(), rect.y() + rect.height(),
                       rect.x() + rect.width(), rect.y() + rect.height())

        p.restore()

        p.save()
        text = fm.elidedText(opt.text, Qt.ElideRight, text_rect.width())
        p.setPen(QPen(palette.color(foregroundrole)))
        p.setFont(opt.font)

        p.drawText(text_rect,
                   int(Qt.AlignVCenter | Qt.AlignLeft) |
                   int(Qt.TextSingleLine),
                   text)

        if not opt.icon.isNull():
            if opt.state & QStyle.State_Enabled:
                mode = QIcon.Normal
            else:
                mode = QIcon.Disabled
            if opt.state & QStyle.State_On:
                state = QIcon.On
            else:
                state = QIcon.Off
            icon_area_rect = icon_area_rect
            icon_rect = QRect(QPoint(0, 0), opt.iconSize)
            icon_rect.moveCenter(icon_area_rect.center())
            opt.icon.paint(p, icon_rect, Qt.AlignCenter, mode, state)
        p.restore()


class _ToolBoxLayout(QVBoxLayout):
    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None
        self.__minimumSize = None  # type: Optional[QSize]
        self.__maximumSize = None  # type: Optional[QSize]
        super().__init__(*args, **kwargs)

    def minimumSize(self):  # type: () -> QSize
        """Reimplemented from `QBoxLayout.minimumSize`."""
        if self.__minimumSize is None:
            msize = super().minimumSize()
            # Extend the minimum size by including the minimum width of
            # hidden widgets (which QBoxLayout ignores), so the minimum
            # width does not depend on the tab open/close state.
            for i in range(self.count()):
                item = self.itemAt(i)
                if item.isEmpty() and item.widget() is not None:
                    msize.setWidth(max(item.widget().minimumWidth(),
                                       msize.width()))
            self.__minimumSize = msize

        return self.__minimumSize

    def maximumSize(self):  # type: () -> QSize
        """Reimplemented from `QBoxLayout.maximumSize`."""
        msize = super().maximumSize()
        # Allow the contents to grow horizontally (expand within the
        # containing scroll area - joining the tab buttons to the
        # right edge), but have a suitable maximum height (displaying an
        # empty area on the bottom if the contents are smaller then the
        # viewport).
        msize.setWidth(QWIDGETSIZE_MAX)
        return msize

    def invalidate(self):  # type: () -> None
        """Reimplemented from `QVBoxLayout.invalidate`."""
        self.__minimumSize = None
        self.__maximumSize = None
        super().invalidate()


class ToolBox(QFrame):
    """
    A tool box widget.
    """
    # Signal emitted when a tab is toggled.
    tabToggled = Signal(int, bool)

    __exclusive = False  # type: bool

    def setExclusive(self, exclusive):  # type: (bool) -> None
        """
        Set exclusive tabs (only one tab can be open at a time).
        """
        if self.__exclusive != exclusive:
            self.__exclusive = exclusive
            self.__tabActionGroup.setExclusive(exclusive)
            checked = self.__tabActionGroup.checkedAction()
            if checked is None:
                # The action group can be out of sync with the actions state
                # when switching between exclusive states.
                actions_checked = [page.action for page in self.__pages
                                   if page.action.isChecked()]
                if actions_checked:
                    checked = actions_checked[0]

            # Trigger/toggle remaining open pages
            if exclusive and checked is not None:
                for page in self.__pages:
                    if checked != page.action and page.action.isChecked():
                        page.action.trigger()

    def exclusive(self):  # type: () -> bool
        """
        Are the tabs in the toolbox exclusive.
        """
        return self.__exclusive

    exclusive_ = Property(bool,
                          fget=exclusive,
                          fset=setExclusive,
                          designable=True,
                          doc="Exclusive tabs")

    def __init__(self, parent=None, **kwargs):
        # type: (Optional[QWidget], Any)-> None
        super().__init__(parent, **kwargs)
        self.__pages = []  # type: List[_ToolBoxPage]
        self.__tabButtonHeight = -1
        self.__tabIconSize = QSize()
        self.__exclusive = False
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        # Scroll area for the contents.
        self.__scrollArea = QScrollArea(
            self, objectName="toolbox-scroll-area",
            sizePolicy=QSizePolicy(QSizePolicy.MinimumExpanding,
                                   QSizePolicy.MinimumExpanding),
            horizontalScrollBarPolicy=Qt.ScrollBarAlwaysOff,
            widgetResizable=True,
        )
        sb = ScrollBar()
        sb.styleChange.connect(self.updateGeometry)
        self.__scrollArea.setVerticalScrollBar(sb)
        self.__scrollArea.setFrameStyle(QScrollArea.NoFrame)

        # A widget with all of the contents.
        # The tabs/contents are placed in the layout inside this widget
        self.__contents = QWidget(self.__scrollArea,
                                  objectName="toolbox-contents")
        self.__contentsLayout = _ToolBoxLayout(
            sizeConstraint=_ToolBoxLayout.SetMinAndMaxSize,
            spacing=0
        )
        self.__contentsLayout.setContentsMargins(0, 0, 0, 0)
        self.__contents.setLayout(self.__contentsLayout)

        self.__scrollArea.setWidget(self.__contents)

        layout.addWidget(self.__scrollArea)

        self.setLayout(layout)
        self.setSizePolicy(QSizePolicy.Fixed,
                           QSizePolicy.MinimumExpanding)

        self.__tabActionGroup = QActionGroup(
            self, objectName="toolbox-tab-action-group",
        )
        self.__tabActionGroup.setExclusive(self.__exclusive)
        self.__actionMapper = QSignalMapper(self)
        self.__actionMapper.mappedObject.connect(self.__onTabActionToggled)

    def setTabButtonHeight(self, height):
        # type: (int) -> None
        """
        Set the tab button height.
        """
        if self.__tabButtonHeight != height:
            self.__tabButtonHeight = height
            for page in self.__pages:
                page.button.setFixedHeight(height)

    def tabButtonHeight(self):
        # type: () -> int
        """
        Return the tab button height.
        """
        return self.__tabButtonHeight

    def setTabIconSize(self, size):
        # type: (QSize) -> None
        """
        Set the tab button icon size.
        """
        if self.__tabIconSize != size:
            self.__tabIconSize = QSize(size)
            for page in self.__pages:
                page.button.setIconSize(size)

    def tabIconSize(self):
        # type: () -> QSize
        """
        Return the tab icon size.
        """
        return QSize(self.__tabIconSize)

    def tabButton(self, index):
        # type: (int) -> QAbstractButton
        """
        Return the tab button at `index`
        """
        return self.__pages[index].button

    def tabAction(self, index):
        # type: (int) -> QAction
        """
        Return open/close action for the tab at `index`.
        """
        return self.__pages[index].action

    def addItem(self, widget, text, icon=QIcon(), toolTip=""):
        # type: (QWidget, str, QIcon, str) -> int
        """
        Append the `widget` in a new tab and return its index.

        Parameters
        ----------
        widget : QWidget
            A widget to be inserted. The toolbox takes ownership
            of the widget.
        text : str
            Name/title of the new tab.
        icon : QIcon
            An icon for the tab button.
        toolTip : str
            Tool tip for the tab button.

        Returns
        -------
        index : int
            Index of the inserted tab
        """
        return self.insertItem(self.count(), widget, text, icon, toolTip)

    def insertItem(self, index, widget, text, icon=QIcon(), toolTip=""):
        # type: (int, QWidget, str, QIcon, str) -> int
        """
        Insert the `widget` in a new tab at position `index`.

        See also
        --------
        ToolBox.addItem
        """
        button = self.createTabButton(widget, text, icon, toolTip)

        self.__contentsLayout.insertWidget(index * 2, button)
        self.__contentsLayout.insertWidget(index * 2 + 1, widget)

        widget.hide()

        page = _ToolBoxPage(index, widget, button.defaultAction(), button)
        self.__pages.insert(index, page)

        # update the indices __pages list
        for i in range(index + 1, self.count()):
            self.__pages[i] = self.__pages[i]._replace(index=i)

        self.__updatePositions()

        # Show (open) the first tab.
        if self.count() == 1 and index == 0:
            page.action.trigger()

        self.__updateSelected()

        self.updateGeometry()
        return index

    def removeItem(self, index):
        # type: (int) -> None
        """
        Remove the widget at `index`.

        Note
        ----
        The widget is hidden but is is not deleted. It is up to the caller to
        delete it.
        """
        self.__contentsLayout.takeAt(2 * index + 1)
        self.__contentsLayout.takeAt(2 * index)
        page = self.__pages.pop(index)

        # Update the page indexes
        for i in range(index, self.count()):
            self.__pages[i] = self.__pages[i]._replace(index=i)

        page.button.deleteLater()

        # Hide the widget and reparent to self
        # This follows QToolBox.removeItem
        page.widget.hide()
        page.widget.setParent(self)

        self.__updatePositions()
        self.__updateSelected()

        self.updateGeometry()

    def count(self):
        # type: () -> int
        """
        Return the number of widgets inserted in the toolbox.
        """
        return len(self.__pages)

    def widget(self, index):
        # type: (int) -> QWidget
        """
        Return the widget at `index`.
        """
        return self.__pages[index].widget

    def createTabButton(self, widget, text, icon=QIcon(), toolTip=""):
        # type: (QWidget, str, QIcon, str) -> QAbstractButton
        """
        Create the tab button for `widget`.
        """
        action = QAction(text, self)
        action.setCheckable(True)

        if icon:
            action.setIcon(icon)

        if toolTip:
            action.setToolTip(toolTip)
        self.__tabActionGroup.addAction(action)
        self.__actionMapper.setMapping(action, action)
        action.toggled.connect(self.__actionMapper.map)

        button = ToolBoxTabButton(self, objectName="toolbox-tab-button")
        button.setDefaultAction(action)
        button.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        button.setSizePolicy(QSizePolicy.Ignored,
                             QSizePolicy.Fixed)

        if self.__tabIconSize.isValid():
            button.setIconSize(self.__tabIconSize)

        if self.__tabButtonHeight > 0:
            button.setFixedHeight(self.__tabButtonHeight)

        return button

    def ensureWidgetVisible(self, child, xmargin=50, ymargin=50):
        # type: (QWidget, int, int) -> None
        """
        Scroll the contents so child widget instance is visible inside
        the viewport.
        """
        self.__scrollArea.ensureWidgetVisible(child, xmargin, ymargin)

    def sizeHint(self):
        # type: () -> QSize
        """
        Reimplemented.
        """
        hint = self.__contentsLayout.sizeHint()

        if self.count():
            # Compute max width of hidden widgets also.
            scroll = self.__scrollArea

            # check if scrollbar is transient
            scrollBar = self.__scrollArea.verticalScrollBar()
            transient = scrollBar.style().styleHint(QStyle.SH_ScrollBar_Transient,
                                                    widget=scrollBar)

            scroll_w = scroll.verticalScrollBar().sizeHint().width() if not transient else 0

            frame_w = self.frameWidth() * 2 + scroll.frameWidth() * 2
            max_w = max([p.widget.sizeHint().width() for p in self.__pages])
            hint = QSize(max(max_w, hint.width()) + scroll_w + frame_w,
                         hint.height())

        return QSize(200, 200).expandedTo(hint)

    def __onTabActionToggled(self, action):
        # type: (QAction) -> None
        page = find(self.__pages, action, key=attrgetter("action"))
        on = action.isChecked()
        page.widget.setVisible(on)
        index = page.index

        if index > 0:
            # Update the `previous` tab buttons style hints
            previous = self.__pages[index - 1].button
            previous.selected = set_flag(
                previous.selected, ToolBoxTabButton.NextIsSelected, on
            )
            previous.update()
        if index < self.count() - 1:
            next = self.__pages[index + 1].button
            next.selected = set_flag(
                next.selected, ToolBoxTabButton.PreviousIsSelected, on
            )
            next.update()

        self.tabToggled.emit(index, on)

        self.__contentsLayout.invalidate()

    def __updateSelected(self):
        # type: () -> None
        """Update the tab buttons selected style flags.
        """
        if self.count() == 0:
            return

        def update(button, next_sel, prev_sel):
            # type: (ToolBoxTabButton, bool, bool) -> None
            button.selected = set_flag(
                button.selected,
                ToolBoxTabButton.NextIsSelected,
                next_sel
            )
            button.selected = set_flag(
                button.selected,
                ToolBoxTabButton.PreviousIsSelected,
                prev_sel
            )
            button.update()

        if self.count() == 1:
            update(self.__pages[0].button, False, False)
        elif self.count() >= 2:
            pages = self.__pages
            for i in range(1, self.count() - 1):
                update(pages[i].button,
                       pages[i + 1].action.isChecked(),
                       pages[i - 1].action.isChecked())

    def __updatePositions(self):
        # type: () -> None
        """Update the tab buttons position style flags.
        """
        if self.count() == 0:
            return
        elif self.count() == 1:
            self.__pages[0].button.position = ToolBoxTabButton.OnlyOneTab
        else:
            self.__pages[0].button.position = ToolBoxTabButton.Beginning
            self.__pages[-1].button.position = ToolBoxTabButton.End
            for p in self.__pages[1:-1]:
                p.button.position = ToolBoxTabButton.Middle

        for p in self.__pages:
            p.button.update()


if typing.TYPE_CHECKING:
    A = typing.TypeVar("A")
    B = typing.TypeVar("B")
    C = typing.TypeVar("C")


def identity(arg):
    return arg


def find(iterable, what, key=identity, predicate=eq):
    # type: (Iterable[A], B, Callable[[A], C], Callable[[C, B], bool]) -> A
    """
    find(iterable, what, [key=None, [predicate=operator.eq]])
    """
    for item in iterable:
        item_key = key(item)
        if predicate(item_key, what):
            return item
    else:
        raise ValueError(what)



"""
Qt Property Bindings (`propertybindings`)
-----------------------------------------


"""

import sys
import ast

from collections import defaultdict
from functools import reduce
from operator import add

from AnyQt.QtCore import QObject, QEvent, QMetaProperty

from AnyQt.QtCore import pyqtSignal as Signal, pyqtSlot as Slot

from typing import Optional, Any, List, Dict, Tuple


def find_meta_property(obj, name):
    # type: (QObject, str) -> QMetaProperty
    """
    Return a named (`name`) `QMetaProperty` of a `QObject` instance `obj`.
    If a property by taht name does not exist raise an AttributeError.

    """
    meta = obj.metaObject()
    index = meta.indexOfProperty(name)
    if index == -1:
        raise AttributeError("%s does no have a property named %r." %
                             (meta.className(), name))

    return meta.property(index)


def find_notifier(obj, name):
    # type: (QObject, str) -> str
    """
    Return the notifier signal name (`str`) for the property of
    `object` (instance of `QObject`).
    """
    prop_meta = find_meta_property(obj, name)
    if not prop_meta.hasNotifySignal():
        raise TypeError("%s does not have a notifier signal." %
                        name)

    notifier = prop_meta.notifySignal()
    name = bytes(notifier.methodSignature()).decode("utf-8").split("(")[0]
    return name


class AbstractBoundProperty(QObject):
    """
    An abstract base class for property bindings.
    """

    changed = Signal([], [object])
    """Emited when the property changes"""

    def __init__(self, obj, propertyName, parent=None):
        # type: (QObject, str, Optional[QObject]) -> None
        super().__init__(parent)

        self.obj = obj  # type: Optional[QObject]
        self.propertyName = propertyName
        self.obj.destroyed.connect(self._on_destroyed)
        self._source = None  # type: Optional[AbstractBoundProperty]

    def set(self, value):
        # type: (Any) -> bool
        """
        Set `value` to the property.
        """
        if self.obj is not None:
            return self.obj.setProperty(self.propertyName, value)
        else:
            return False

    def get(self):
        # type: () -> Any
        """
        Return the property value.
        """
        if self.obj is not None:
            return self.obj.property(self.propertyName)
        else:
            return None

    @Slot()
    def notifyChanged(self):
        """
        Notify the binding of a change in the property value.
        The default implementation emits the `changed` signals.
        """
        val = self.get()
        self.changed.emit()
        self.changed[object].emit(val)

    def _on_destroyed(self):
        # type: () -> None
        self.obj = None

    def bindTo(self, source):
        # type: (AbstractBoundProperty) -> None
        """
        Bind this property to `source` (instance of `AbstractBoundProperty`).
        """
        if self._source != source:
            if self._source:
                self.unbind()

            self._source = source

            source.changed.connect(self.update)
            source.destroyed.connect(self.unbind)
            self.set(source.get())
            self.notifyChanged()

    def unbind(self):
        # type: () -> None
        """
        Unbind the currently bound property (set with `bindTo`).
        """
        if self._source is not None:
            self._source.destroyed.disconnect(self.unbind)
            self._source.changed.disconnect(self.update)
            self._source = None

    def update(self):
        # type: () -> None
        """
        Update the property value from `source` property (`bindTo`).
        """
        if self._source:
            source_val = self._source.get()
            curr_val = self.get()
            if source_val != curr_val:
                self.set(source_val)

    def reset(self):
        """
        Reset the property if possible.
        """
        raise NotImplementedError


class PropertyBindingExpr(AbstractBoundProperty):
    def __init__(self, expression, globals={}, locals={}, parent=None):
        # skip the AbstractBoundProperty's __init__. Despite its name it is
        # not an abstract.
        QObject.__init__(self, parent)

        self.ast = ast.parse(expression, mode="eval")
        self.code = compile(self.ast, "<unknown>", "eval")

        self.expression = expression
        self.globals = dict(globals)
        self.locals = dict(locals)
        self._sources = {}  # type: Dict[str, AbstractBoundProperty]

        names = self.code.co_names
        for name in names:
            v = locals.get(name, globals.get(name))
            if isinstance(v, AbstractBoundProperty):
                self._sources[name] = v
                v.changed.connect(self.notifyChanged)
                v.destroyed.connect(self._on_destroyed)

    def sources(self):
        # type: () -> List[AbstractBoundProperty]
        """Return all source property bindings appearing in the
        expression namespace.
        """
        return list(self._sources.values())

    def set(self, value):
        raise NotImplementedError("Cannot set a value of an expression")

    def get(self):
        # type: () -> Any
        locals = dict(self.locals)
        locals.update(dict((name, source.get())
                           for name, source in self._sources.items()))
        try:
            value = eval(self.code, self.globals, locals)
        except Exception:
            raise
        return value

    def bindTo(self, source):
        raise NotImplementedError("Cannot bind an expression")

    def _on_destroyed(self):
        # type: () -> None
        source = self.sender()
        for name, obj in list(self._sources.items()):
            if obj is source:
                del self._sources[name]


class PropertyBinding(AbstractBoundProperty):
    """
    A Property binding of a QObject's property registered with Qt's
    meta class object system.

    """
    def __init__(self, obj, propertyName, notifier=None, parent=None):
        super().__init__(obj, propertyName, parent)

        if notifier is None:
            notifier = find_notifier(obj, propertyName)

        if notifier is not None:
            signal = getattr(obj, notifier)
            signal.connect(self.notifyChanged)
        else:
            signal = None

        self.notifierSignal = signal

    def _on_destroyed(self):
        self.notifierSignal = None
        super()._on_destroyed()

    def reset(self):
        meta_prop = find_meta_property(self.obj, self.propertyName)
        if meta_prop.isResettable():
            meta_prop.reset(self.obj)
        else:
            return super().reset()


class DynamicPropertyBinding(AbstractBoundProperty):
    """
    A Property binding of a QObject's dynamic property.
    """
    def __init__(self, obj, propertyName, parent=None):
        super().__init__(obj, propertyName, parent)

        obj.installEventFilter(self)

    def eventFilter(self, obj, event):
        if obj is self.obj and event.type() == QEvent.DynamicPropertyChange:
            if event.propertyName() == self.propertyName:
                self.notifyChanged()

        return super().eventFilter(obj, event)


class BindingManager(QObject):
    AutoSubmit = 0
    ManualSubmit = 1

    # Note: This should also apply to Gnome
    Default = 0 if sys.platform == "darwin" else 1

    def __init__(self, parent=None, submitPolicy=Default):
        super().__init__(parent)
        self._bindings = defaultdict(list)
        self._modified = set()
        self.__submitPolicy = submitPolicy

    def setSubmitPolicy(self, policy):
        if self.__submitPolicy != policy:
            self.__submitPolicy = policy
            if policy == BindingManager.AutoSubmit:
                self.commit()

    def submitPolicy(self):
        return self.__submitPolicy

    def bind(self, target, source):
        if isinstance(target, tuple):
            target = binding_for(*target + (self, ))

        if source is None:
            return UnboundBindingWrapper(target, self)

        else:
            if isinstance(source, tuple):
                source = binding_for(*source + (self,))

            source.changed.connect(self.__on_changed)
            self._bindings[source].append((target, source))
            self.__on_changed(source)

            return None

    def bindings(self):
        """Return (target, source) binding tuples.
        """
        return reduce(add, self._bindings.items(), [])

    def commit(self):
        self.__update()

    def __on_changed(self, sender=None):
        if sender is None:
            sender = self.sender()
        self._modified.add(sender)
        if self.__submitPolicy == BindingManager.AutoSubmit:
            self.__update()

    def __update(self):
        for modified in list(self._modified):
            self._modified.remove(modified)
            for target, source in self._bindings.get(modified, []):
                target.set(source.get())


class UnboundBindingWrapper(object):
    def __init__(self, target, manager):
        self.target = target
        self.manager = manager
        self.__source = None

    def to(self, source):
        if self.__source is None:
            if isinstance(source, tuple):
                source = binding_for(*source + (self.manager,))

            self.manager.bind(self.target, source)
            self.__source = source
        else:
            raise ValueError("Can only call 'to' once.")


def binding_for(obj, name, parent=None):
    """
    Return a suitable binding for property `name` of an `obj`.
    Currently only supports PropertyBinding and DynamicPropertyBinding.

    """
    if isinstance(obj, QObject):
        meta = obj.metaObject()
        index = meta.indexOfProperty(name)
        if index == -1:
            boundprop = DynamicPropertyBinding(obj, name, parent)
        else:
            boundprop = PropertyBinding(obj, name, parent)
    else:
        raise TypeError
    return boundprop



import time
import gc
import weakref
from types import SimpleNamespace
from typing import List, Optional, Callable

import unittest
from AnyQt.QtCore import QCoreApplication, QThread, QObject, Signal, QEvent

from concurrent.futures.thread import ThreadPoolExecutor

from AnyQt.QtTest import QSignalSpy, QTest

from ..qinvoke import qinvoke, connect_with_context


class CoreAppTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.app = QCoreApplication.instance()
        if self.app is None:
            self.app = QCoreApplication([])
        super().setUp()

    def tearDown(self) -> None:
        self.app = None
        super().tearDown()


def delete(qobj: QObject):
    assert qobj.thread() is QThread.currentThread()
    spy = QSignalSpy(qobj.destroyed)
    qobj.deleteLater()
    QCoreApplication.sendPostedEvents(qobj, QEvent.DeferredDelete)
    assert len(spy) == 1


class TestMethodinvoke(CoreAppTestCase):
    def test_qinvoke(self):
        executor = ThreadPoolExecutor()
        state = [None, None]  # type: List[Optional[QThread]]

        class StateSetter(QObject):
            didsetstate = Signal()

            def set_state(self, value: QThread) -> None:
                state[0] = value
                state[1] = QThread.currentThread()
                self.didsetstate.emit()

        def func(callback):  # type: (Callable[[QThread], None]) -> None
            callback(QThread.currentThread())

        obj = StateSetter()
        spy = QSignalSpy(obj.didsetstate)
        f = executor.submit(
            func, qinvoke(obj.set_state, context=obj)
        )
        self.assertTrue(spy.wait())

        self.assertIs(state[1], QThread.currentThread(),
                      "set_state was called from the wrong thread")

        self.assertIsNot(state[0], QThread.currentThread(),
                         "set_state was invoked in the main thread")

        # test that disconnect works 'atomically' w.r.t. event loop
        spy = QSignalSpy(obj.didsetstate)
        callback = qinvoke(obj.set_state, context=obj)
        f = executor.submit(func, callback)
        f.result()
        time.sleep(0.01)  # wait for finish
        callback.disconnect()  # type: ignore
        self.assertFalse(spy.wait(100), "")
        self.assertSequenceEqual(spy, [])

        executor.shutdown(wait=True)

    def test_qinvoke_context_delete(self):
        executor = ThreadPoolExecutor(max_workers=1)
        context = QObject()
        isdeleted = False
        lastindex = -1

        def mark_deleted():
            nonlocal isdeleted
            isdeleted = True

        context.destroyed.connect(mark_deleted)

        def func(i):
            nonlocal isdeleted
            nonlocal lastindex
            lastindex = i
            self.assertFalse(isdeleted)
            self.assertIs(context.thread(), self.app.thread())

        callback = qinvoke(func, context=context)

        _ = executor.map(callback, range(1000))

        while lastindex < 0:
            QTest.qWait(10)
        assert lastindex >= 0
        delete(context)
        assert isdeleted
        lasti = lastindex
        QTest.qWait(50)
        assert lasti == lastindex
        executor.shutdown()

    def test_qinvoke_as_decorator(self):
        context = QObject()

        @qinvoke(context=context)
        def f(name):
            context.setObjectName(name)

        spy = QSignalSpy(context.objectNameChanged)
        f("name")
        self.assertTrue(spy.wait(500))

    def test_errors(self):
        with self.assertRaises(TypeError):
            qinvoke(lambda: None)

        with self.assertRaises(TypeError):
            qinvoke()


class TestConnectWithContext(CoreAppTestCase):
    def test_connect_with_context(self):
        state = SimpleNamespace(th=None, greeting=None)
        target = QObject()

        def in_target_context(greeting: str):
            state.th = QThread.currentThread()
            state.greeting = greeting
            target.setObjectName(greeting)

        executor = ThreadPoolExecutor()
        emiter = QObject()
        connect_with_context(
            emiter.objectNameChanged,
            target,
            in_target_context,
        )

        def run():
            emiter.objectNameChanged.emit("hello")

        spy = QSignalSpy(target.objectNameChanged)
        executor.submit(run)
        self.assertTrue(spy.wait())
        self.assertIs(target.thread(), state.th)

        # delete target context (and connection)
        delete(target)
        state.th = None
        state.greeting = None

        executor.map(run, range(10))
        executor.shutdown(wait=True)
        QTest.qWait(10)
        self.assertIsNone(state.th)
        self.assertIsNone(state.greeting)

    def test_gc_captured_context(self):
        emitter = QObject()
        context = QObject()
        ref = weakref.ref(context)
        connect_with_context(
            emitter.objectNameChanged,
            context,
            lambda name: context.setObjectName(name)  # captures context
        )
        emitter.setObjectName("AA")
        self.assertEqual(context.objectName(), emitter.objectName())
        del context  # This needs to clear captured context in connected func
        self.assertIsNone(ref())
        emitter.setObjectName("BB")

    def test_gc_captured_context_cycle(self):
        emitter = QObject()
        context = QObject()
        context.cycle = context
        ref = weakref.ref(context)
        connect_with_context(
            emitter.objectNameChanged,
            context,
            lambda name: context.setObjectName(name)  # captures context
        )
        emitter.setObjectName("AA")
        self.assertEqual(context.objectName(), emitter.objectName())
        delete(context)
        del context  # This needs to clear captured context in connected func
        gc.collect()
        self.assertIsNone(ref())
        emitter.setObjectName("BB")



"""
Tests for scheme document.
"""
import sys
import unittest
from unittest import mock
from typing import Iterable

from AnyQt.QtCore import Qt, QPoint, QMimeData
from AnyQt.QtGui import QPainterPath
from AnyQt.QtWidgets import (
    QGraphicsWidget, QAction, QApplication, QMenu, QWidget
)
from AnyQt.QtTest import QSignalSpy, QTest

from .. import commands
from ..schemeedit import SchemeEditWidget, SaveWindowGroup
from ..interactions import (
    DropHandler, PluginDropHandler, NodeFromMimeDataDropHandler, EntryPoint
)
from ...canvas import items
from ...scheme import Scheme, SchemeNode, SchemeLink, SchemeTextAnnotation, \
                      SchemeArrowAnnotation
from ...registry.tests import small_testing_registry
from ...gui.test import QAppTestCase, mouseMove, dragDrop, dragEnterLeave, \
    contextMenu
from ...utils import findf
from ...scheme.tests.test_widgetmanager import TestingWidgetManager


def action_by_name(actions, name):
    # type: (Iterable[QAction], str) -> QAction
    for a in actions:
        if a.objectName() == name:
            return a
    raise LookupError(name)


class TestSchemeEdit(QAppTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.reg = small_testing_registry()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        del cls.reg

    def setUp(self):
        super().setUp()
        self.w = SchemeEditWidget()
        self.w.setScheme(Scheme())
        self.w.setRegistry(self.reg)
        self.w.resize(300, 300)

    def tearDown(self):
        del self.w
        super().tearDown()

    def test_schemeedit(self):
        reg = self.reg
        w = self.w

        scheme = Scheme()
        w.setScheme(scheme)

        self.assertIs(w.scheme(), scheme)
        self.assertFalse(w.isModified())

        scheme = Scheme()
        w.setScheme(scheme)

        self.assertIs(w.scheme(), scheme)
        self.assertFalse(w.isModified())

        w.show()

        one_desc = reg.widget("one")
        negate_desc = reg.widget("negate")

        node_list = []
        link_list = []
        annot_list = []

        scheme.node_added.connect(node_list.append)
        scheme.node_removed.connect(node_list.remove)

        scheme.link_added.connect(link_list.append)
        scheme.link_removed.connect(link_list.remove)

        scheme.annotation_added.connect(annot_list.append)
        scheme.annotation_removed.connect(annot_list.remove)

        node = SchemeNode(one_desc, title="title1", position=(100, 100))
        w.addNode(node)

        self.assertSequenceEqual(node_list, [node])
        self.assertSequenceEqual(scheme.nodes, node_list)

        self.assertTrue(w.isModified())

        stack = w.undoStack()
        stack.undo()

        self.assertSequenceEqual(node_list, [])
        self.assertSequenceEqual(scheme.nodes, node_list)
        self.assertTrue(not w.isModified())

        stack.redo()

        node1 = SchemeNode(negate_desc, title="title2", position=(300, 100))
        w.addNode(node1)

        self.assertSequenceEqual(node_list, [node, node1])
        self.assertSequenceEqual(scheme.nodes, node_list)
        self.assertTrue(w.isModified())

        link = SchemeLink(node, "value", node1, "value")
        w.addLink(link)

        self.assertSequenceEqual(link_list, [link])

        stack.undo()
        stack.undo()

        stack.redo()
        stack.redo()

        w.removeNode(node1)

        self.assertSequenceEqual(link_list, [])
        self.assertSequenceEqual(node_list, [node])

        stack.undo()

        self.assertSequenceEqual(link_list, [link])
        self.assertSequenceEqual(node_list, [node, node1])

        spy = QSignalSpy(node.title_changed)
        w.renameNode(node, "foo bar")
        self.assertSequenceEqual(list(spy), [["foo bar"]])
        self.assertTrue(w.isModified())
        stack.undo()
        self.assertSequenceEqual(list(spy), [["foo bar"], ["title1"]])

        w.removeLink(link)

        self.assertSequenceEqual(link_list, [])

        stack.undo()

        self.assertSequenceEqual(link_list, [link])

        annotation = SchemeTextAnnotation((200, 300, 50, 20), "text")
        w.addAnnotation(annotation)
        self.assertSequenceEqual(annot_list, [annotation])

        stack.undo()
        self.assertSequenceEqual(annot_list, [])

        stack.redo()
        self.assertSequenceEqual(annot_list, [annotation])

        w.removeAnnotation(annotation)
        self.assertSequenceEqual(annot_list, [])
        stack.undo()
        self.assertSequenceEqual(annot_list, [annotation])

        self.assertTrue(w.isModified())
        self.assertFalse(stack.isClean())
        w.setModified(False)

        self.assertFalse(w.isModified())
        self.assertTrue(stack.isClean())
        w.setModified(True)
        self.assertTrue(w.isModified())

    def test_modified(self):
        node = SchemeNode(
            self.reg.widget("one"), title="title1", position=(100, 100))
        self.w.addNode(node)
        self.assertTrue(self.w.isModified())
        self.w.setModified(False)
        self.assertFalse(self.w.isModified())
        self.w.setTitle("Title")
        self.assertTrue(self.w.isModified())
        self.w.setDescription("AAA")
        self.assertTrue(self.w.isModified())
        undo = self.w.undoStack()
        undo.undo()
        undo.undo()
        self.assertFalse(self.w.isModified())

    def test_teardown(self):
        w = self.w
        w.undoStack().isClean()
        new = Scheme()
        w.setScheme(new)

    def test_actions(self):
        w = self.w
        actions = w.toolbarActions()

        action_by_name(actions, "action-zoom-in").trigger()
        action_by_name(actions, "action-zoom-out").trigger()
        action_by_name(actions, "action-zoom-reset").trigger()

    def test_node_rename(self):
        w = self.w
        view = w.view()
        node = SchemeNode(self.reg.widget("one"), title="A")
        w.addNode(node)
        w.editNodeTitle(node)
        # simulate editing
        QTest.keyClicks(view.viewport(), "BB")
        QTest.keyClick(view.viewport(), Qt.Key_Enter)
        self.assertEqual(node.title, "BB")
        # last undo command must be rename command
        undo = w.undoStack()
        command = undo.command(undo.count() - 1)
        self.assertIsInstance(command, commands.RenameNodeCommand)

    @unittest.skipUnless(sys.platform == "darwin", "macos only")
    def test_node_rename_click_selected(self):
        w = self.w
        scene = w.scene()
        view = w.view()
        w.show()
        w.raise_()
        w.activateWindow()
        node = SchemeNode(self.reg.widget("one"), title="A")
        w.addNode(node)
        w.selectAll()
        item = scene.item_for_node(node)
        assert isinstance(item, items.NodeItem)
        point = item.captionTextItem.boundingRect().center()
        point = item.captionTextItem.mapToScene(point)
        point = view.mapFromScene(point)
        QTest.mouseClick(view.viewport(), Qt.LeftButton, Qt.NoModifier, point)
        self.assertTrue(item.captionTextItem.isEditing())
        contextMenu(view.viewport(), point)

    def test_arrow_annotation_action(self):
        w = self.w
        workflow = w.scheme()
        workflow.clear()
        view = w.view()
        actions = w.toolbarActions()
        action_by_name(actions, "new-arrow-action").trigger()
        QTest.mousePress(view.viewport(), Qt.LeftButton, pos=QPoint(50, 50))
        mouseMove(view.viewport(), Qt.LeftButton, pos=QPoint(100, 100))
        QTest.mouseRelease(view.viewport(), Qt.LeftButton, pos=QPoint(100, 100))
        self.assertEqual(len(workflow.annotations), 1)
        self.assertIsInstance(workflow.annotations[0], SchemeArrowAnnotation)

    def test_arrow_annotation_action_cancel(self):
        w = self.w
        workflow = w.scheme()
        view = w.view()
        actions = w.toolbarActions()
        action = action_by_name(actions, "new-arrow-action")
        action.trigger()
        self.assertTrue(action.isChecked())
        # cancel immediately after activating
        QTest.keyClick(view.viewport(), Qt.Key_Escape)
        self.assertFalse(action.isChecked())
        action.trigger()
        # cancel after mouse press and drag
        QTest.mousePress(view.viewport(), Qt.LeftButton, pos=QPoint(50, 50))
        mouseMove(view.viewport(), Qt.LeftButton, pos=QPoint(100, 100))
        QTest.keyClick(view.viewport(), Qt.Key_Escape)
        self.assertFalse(action.isChecked())
        self.assertEqual(workflow.annotations, [])

    def test_text_annotation_action(self):
        w = self.w
        workflow = w.scheme()
        workflow.clear()
        view = w.view()
        actions = w.toolbarActions()
        action_by_name(actions, "new-text-action").trigger()
        QTest.mousePress(view.viewport(), Qt.LeftButton, pos=QPoint(50, 50))
        mouseMove(view.viewport(), Qt.LeftButton, pos=QPoint(100, 100))
        QTest.mouseRelease(view.viewport(), Qt.LeftButton, pos=QPoint(100, 100))
        # need to steal focus from the item for it to be commited.
        w.scene().setFocusItem(None)

        self.assertEqual(len(workflow.annotations), 1)
        self.assertIsInstance(workflow.annotations[0], SchemeTextAnnotation)

    def test_text_annotation_action_cancel(self):
        w = self.w
        workflow = w.scheme()
        view = w.view()
        actions = w.toolbarActions()
        action = action_by_name(actions, "new-text-action")
        action.trigger()
        self.assertTrue(action.isChecked())
        # cancel immediately after activating
        QTest.keyClick(view.viewport(), Qt.Key_Escape)
        self.assertFalse(action.isChecked())
        action.trigger()
        # cancel after mouse press and drag
        QTest.mousePress(view.viewport(), Qt.LeftButton, pos=QPoint(50, 50))
        mouseMove(view.viewport(), Qt.LeftButton, pos=QPoint(100, 100))
        QTest.keyClick(view.viewport(), Qt.Key_Escape)
        self.assertFalse(action.isChecked())
        w.scene().setFocusItem(None)
        self.assertEqual(workflow.annotations, [])

    def test_path(self):
        w = self.w
        spy = QSignalSpy(w.pathChanged)
        self.w.setPath("/dev/null")
        self.assertSequenceEqual(list(spy), [["/dev/null"]])

    def test_ensure_visible(self):
        w = self.w
        node = SchemeNode(
            self.reg.widget("one"), title="title1", position=(10000, 100))
        self.w.addNode(node)
        w.setFixedSize(300, 300)
        w.show()
        assert QTest.qWaitForWindowExposed(w, 500)
        w.ensureVisible(node)
        view = w.view()
        viewrect = view.mapToScene(view.viewport().geometry()).boundingRect()
        self.assertTrue(viewrect.contains(10000., 100.))

    def test_select(self):
        w = self.w
        self.setup_test_workflow(w.scheme())
        w.selectAll()
        self.assertSequenceEqual(
            w.selectedNodes(), w.scheme().nodes)
        self.assertSequenceEqual(
            w.selectedAnnotations(), w.scheme().annotations)
        self.assertSequenceEqual(
            w.selectedLinks(), w.scheme().links)
        w.removeSelected()
        self.assertEqual(w.scheme().nodes, [])
        self.assertEqual(w.scheme().annotations, [])
        self.assertEqual(w.scheme().links, [])

    def test_select_remove_link(self):
        def link_curve(link: SchemeLink) -> QPainterPath:
            item = scene.item_for_link(link)  # type: items.LinkItem
            path = item.curveItem.curvePath()
            return item.mapToScene(path)
        w = self.w
        workflow = self.setup_test_workflow(w.scheme())
        w.alignToGrid()
        scene, view = w.scene(), w.view()
        link = workflow.links[0]
        path = link_curve(link)
        p = path.pointAtPercent(0.5)
        QTest.mouseClick(view.viewport(), Qt.LeftButton, pos=view.mapFromScene(p))
        self.assertSequenceEqual(w.selectedLinks(), [link])
        w.removeSelected()
        self.assertSequenceEqual(w.selectedLinks(), [])
        self.assertTrue(link not in workflow.links)

    def test_open_selected(self):
        w = self.w
        w.setScheme(self.setup_test_workflow())
        w.selectAll()
        w.openSelected()

    def test_insert_node_on_link(self):
        w = self.w
        workflow = self.setup_test_workflow(w.scheme())
        neg = SchemeNode(self.reg.widget("negate"))
        target = workflow.links[0]
        spyrem = QSignalSpy(workflow.link_removed)
        spyadd = QSignalSpy(workflow.link_added)
        w.insertNode(neg, target)
        self.assertEqual(workflow.nodes[-1], neg)

        self.assertSequenceEqual(list(spyrem), [[target]])
        self.assertEqual(len(spyadd), 2)
        w.undoStack().undo()

    def test_align_to_grid(self):
        w = self.w
        self.setup_test_workflow(w.scheme())
        w.alignToGrid()

    def test_activate_node(self):
        w = self.w
        workflow = self.setup_test_workflow()
        w.setScheme(workflow)

        view, scene = w.view(), w.scene()
        item = scene.item_for_node(workflow.nodes[0])  # type: QGraphicsWidget
        item.setSelected(True)
        item.setFocus(Qt.OtherFocusReason)
        self.assertIs(w.focusNode(), workflow.nodes[0])
        item.activated.emit()

    def test_duplicate(self):
        w = self.w
        workflow = self.setup_test_workflow()
        w.setScheme(workflow)
        w.selectAll()
        nnodes, nlinks = len(workflow.nodes), len(workflow.links)
        a = action_by_name(w.actions(), "duplicate-action")
        a.trigger()
        self.assertEqual(len(workflow.nodes), 2 * nnodes)
        self.assertEqual(len(workflow.links), 2 * nlinks)

    def test_copy_paste(self):
        w = self.w
        workflow = self.setup_test_workflow()
        w.setRegistry(self.reg)
        w.setScheme(workflow)
        w.selectAll()
        nnodes, nlinks = len(workflow.nodes), len(workflow.links)
        ca = action_by_name(w.actions(), "copy-action")
        cp = action_by_name(w.actions(), "paste-action")
        cb = QApplication.clipboard()
        spy = QSignalSpy(cb.dataChanged)
        ca.trigger()
        if not len(spy):
            self.assertTrue(spy.wait())
        self.assertEqual(len(spy), 1)
        cp.trigger()
        self.assertEqual(len(workflow.nodes), 2 * nnodes)
        self.assertEqual(len(workflow.links), 2 * nlinks)

        w1 = SchemeEditWidget()
        w1.setRegistry(self.reg)
        w1.setScheme((Scheme()))
        cp = action_by_name(w1.actions(), "paste-action")
        self.assertTrue(cp.isEnabled())
        cp.trigger()
        wf1 = w1.scheme()
        self.assertEqual(len(wf1.nodes), nnodes)
        self.assertEqual(len(wf1.links), nlinks)

    def test_redo_remove_preserves_order(self):
        w = self.w
        workflow = self.setup_test_workflow()
        w.setRegistry(self.reg)
        w.setScheme(workflow)
        undo = w.undoStack()
        links = workflow.links
        nodes = workflow.nodes
        annotations = workflow.annotations
        assert len(links) > 2
        w.removeLink(links[1])
        self.assertSequenceEqual(links[:1] + links[2:], workflow.links)
        undo.undo()
        self.assertSequenceEqual(links, workflow.links)
        # find add node that has multiple in/out links
        node = findf(workflow.nodes, lambda n: n.title == "add")
        w.removeNode(node)
        undo.undo()
        self.assertSequenceEqual(links, workflow.links)
        self.assertSequenceEqual(nodes, workflow.nodes)

        w.removeAnnotation(annotations[0])
        self.assertSequenceEqual(annotations[1:], workflow.annotations)
        undo.undo()
        self.assertSequenceEqual(annotations, workflow.annotations)

    def test_window_groups(self):
        w = self.w
        workflow = self.setup_test_workflow()
        workflow.set_window_group_presets([
            Scheme.WindowGroup("G1", False, [(workflow.nodes[0], b'\xff\x00')]),
            Scheme.WindowGroup("G2", True, [(workflow.nodes[0], b'\xff\x00')]),
        ])
        manager = TestingWidgetManager()
        workflow.widget_manager = manager
        with mock.patch.object(manager, "activate_window_group") as m:
            w.setScheme(workflow)
            w.activateDefaultWindowGroup()
            m.assert_called_once_with(workflow.window_group_presets()[1])

        a = w.findChild(QAction, "window-groups-save-action")
        with mock.patch.object(
                workflow, "set_window_group_presets",
                wraps=workflow.set_window_group_presets
        ) as m:
            a.trigger()
            dlg = w.findChild(SaveWindowGroup)
            dlg.accept()
            m.assert_called_once()

        with mock.patch.object(
                workflow, "set_window_group_presets",
                wraps=workflow.set_window_group_presets
        ) as m:
            w.undoStack().undo()
            m.assert_called_once()

        with mock.patch.object(
                workflow, "set_window_group_presets",
                wraps=workflow.set_window_group_presets
        ) as m:
            a = w.findChild(QAction, "window-groups-clear-action")
            a.trigger()
            m.assert_called_once_with([])
        workflow.clear()

    def test_drop_event(self):
        w = self.w
        w.setRegistry(self.reg)
        workflow = w.scheme()
        desc = self.reg.widget("one")
        viewport = w.view().viewport()
        mime = QMimeData()
        mime.setData(
            "application/vnd.orange-canvas.registry.qualified-name",
            desc.qualified_name.encode("utf-8")
        )

        self.assertTrue(dragDrop(viewport, mime, QPoint(10, 10)))

        self.assertEqual(len(workflow.nodes), 1)
        self.assertEqual(workflow.nodes[0].description, desc)

        dragEnterLeave(viewport, mime)

        self.assertEqual(len(workflow.nodes), 1)

    def test_drag_drop(self):
        w = self.w
        w.setRegistry(self.reg)
        handler = TestDropHandler()
        w.setDropHandlers([handler])
        viewport = w.view().viewport()
        mime = QMimeData()
        mime.setData(handler.format_, b'abc')

        dragDrop(viewport, mime, QPoint(10, 10))

        self.assertEqual(handler.doDrop_calls, 1)
        self.assertGreaterEqual(handler.accepts_calls, 1)
        self.assertIsNone(w._userInteractionHandler())

        handler.accepts_calls = 0
        handler.doDrop_calls = 0
        mime = QMimeData()
        mime.setData("application/prs.do-not-accept-this", b'abc')

        dragDrop(viewport, mime, QPoint(10, 10))

        self.assertGreaterEqual(handler.accepts_calls, 1)
        self.assertEqual(handler.doDrop_calls, 0)
        self.assertIsNone(w._userInteractionHandler())

        dragEnterLeave(viewport, mime, QPoint(10, 10))

        self.assertIsNone(w._userInteractionHandler())

    @mock.patch.object(
        PluginDropHandler, "iterEntryPoints",
        # "pkg_resources.WorkingSet.iter_entry_points",
        lambda _: [
            EntryPoint(
                "AA", f"{__name__}:TestDropHandler", "aa"
            ),
            EntryPoint(
                "BB", f"{__name__}:TestNodeFromMimeData", "aa"
            )
        ]
    )
    def test_plugin_drag_drop(self):
        handler = PluginDropHandler()
        w = self.w
        w.setRegistry(self.reg)
        w.setDropHandlers([handler])
        workflow = w.scheme()
        viewport = w.view().viewport()
        # Test empty handler
        mime = QMimeData()
        mime.setData(TestDropHandler.format_, b'abc')

        dragDrop(viewport, mime, QPoint(10, 10))

        self.assertIsNone(w._userInteractionHandler())

        # test create node handler
        mime = QMimeData()
        mime.setData(TestNodeFromMimeData.format_, b'abc')

        dragDrop(viewport, mime, QPoint(10, 10))

        self.assertIsNone(w._userInteractionHandler())
        self.assertEqual(len(workflow.nodes), 1)
        self.assertEqual(workflow.nodes[0].description.name, "one")
        self.assertEqual(workflow.nodes[0].properties, {"a": "from drop"})

        workflow.clear()

        # Test both simultaneously (menu for selection)
        mime = QMimeData()
        mime.setData(TestDropHandler.format_, b'abc')
        mime.setData(TestNodeFromMimeData.format_, b'abc')

        def exec(self, *args):
            return action_by_name(self.actions(), "-pick-me")

        # intercept QMenu.exec, force select the TestNodeFromMimeData handler
        with mock.patch.object(QMenu, "exec", exec):
            dragDrop(viewport, mime, QPoint(10, 10))

        self.assertEqual(len(workflow.nodes), 1)
        self.assertEqual(workflow.nodes[0].description.name, "one")
        self.assertEqual(workflow.nodes[0].properties, {"a": "from drop"})

    def test_activate_drop_node(self):
        class NodeFromMimeData(TestNodeFromMimeData):
            def shouldActivateNode(self) -> bool:
                self.shouldActivateNode_called += 1
                return True
            shouldActivateNode_called = 0

            def activateNode(self, document: 'SchemeEditWidget', node: 'Node',
                             widget: 'QWidget') -> None:
                self.activateNode_called += 1
                super().activateNode(document, node, widget)
                widget.didActivate = True
            activateNode_called = 0

        w = self.w
        viewport = w.view().viewport()
        workflow = Scheme()
        wm = workflow.widget_manager = TestingWidgetManager()
        wm.set_creation_policy(TestingWidgetManager.Immediate)
        wm.set_workflow(workflow)
        w.setScheme(workflow)
        handler = NodeFromMimeData()
        w.setDropHandlers([handler])
        mime = QMimeData()
        mime.setData(TestNodeFromMimeData.format_, b'abc')
        record = []
        wm.widget_for_node_added.connect(
            lambda obj, widget: record.append((obj, widget))
        )
        dragDrop(viewport, mime, QPoint(10, 10))
        self.assertEqual(len(record), 1)
        self.assertGreaterEqual(handler.shouldActivateNode_called, 1)
        self.assertGreaterEqual(handler.activateNode_called, 1)
        _, widget = record[0]
        self.assertTrue(widget.didActivate)
        workflow.clear()

    @classmethod
    def setup_test_workflow(cls, scheme=None):
        # type: (Scheme) -> Scheme
        if scheme is None:
            scheme = Scheme()
        reg = cls.reg

        zero_desc = reg.widget("zero")
        one_desc = reg.widget("one")
        add_desc = reg.widget("add")
        negate = reg.widget("negate")

        zero_node = SchemeNode(zero_desc)
        one_node = SchemeNode(one_desc)
        add_node = SchemeNode(add_desc)
        negate_node = SchemeNode(negate)

        scheme.add_node(zero_node)
        scheme.add_node(one_node)
        scheme.add_node(add_node)
        scheme.add_node(negate_node)

        scheme.add_link(SchemeLink(zero_node, "value", add_node, "left"))
        scheme.add_link(SchemeLink(one_node, "value", add_node, "right"))
        scheme.add_link(SchemeLink(add_node, "result", negate_node, "value"))

        scheme.add_annotation(SchemeArrowAnnotation((0, 0), (10, 10)))
        scheme.add_annotation(SchemeTextAnnotation((0, 100, 200, 200), "$$"))
        return scheme


class TestDropHandler(DropHandler):
    format_ = "application/prs.test"
    accepts_calls = 0
    doDrop_calls = 0

    def accepts(self, document, event) -> bool:
        self.accepts_calls += 1
        return event.mimeData().hasFormat(self.format_)

    def doDrop(self, document, event) -> bool:
        self.doDrop_calls += 1
        return event.mimeData().hasFormat(self.format_)


class TestNodeFromMimeData(NodeFromMimeDataDropHandler):
    format_ = "application/prs.one"

    def qualifiedName(self) -> str:
        return "one"

    def canDropMimeData(self, document, data: 'QMimeData') -> bool:
        return data.hasFormat(self.format_)

    def parametersFromMimeData(self, document, data: 'QMimeData') -> 'Dict[str, Any]':
        return {"a": "from drop"}

    def actionFromDropEvent(
            self, document: 'SchemeEditWidget', event: 'QGraphicsSceneDragDropEvent'
    ) -> QAction:
        a = super().actionFromDropEvent(document, event)
        a.setObjectName("-pick-me")
        return a



import sys
import math
from itertools import islice, count
from operator import itemgetter

import typing
from typing import List, Iterable, Optional, Callable, Any, Union, Tuple

from AnyQt.QtCore import QPointF, QLineF
from AnyQt.QtGui import (
    QColor, QRadialGradient, QPainterPathStroker, QPainterPath, QPen
)

from AnyQt.QtWidgets import QGraphicsItem

if typing.TYPE_CHECKING:
    T = typing.TypeVar("T")
    A = typing.TypeVar("A")
    B = typing.TypeVar("B")
    C = typing.TypeVar("C")


def composition(f, g):
    # type: (Callable[[A], B], Callable[[B], C]) -> Callable[[A], C]
    """
    Return a composition of two functions.
    """
    def fg(arg):  # type: (A) -> C
        return g(f(arg))
    return fg


def argsort(iterable, key=None, reverse=False):
    # type: (Iterable[T], Optional[Callable[[T], Any]], bool) -> List[int]
    """
    Return indices that sort elements of iterable in ascending order.

    A custom key function can be supplied to customize the sort order, and the
    reverse flag can be set to request the result in descending order.

    Parameters
    ----------
    iterable : Iterable[T]
    key : Callable[[T], Any]
    reverse : bool

    Returns
    -------
    indices : List[int]
    """
    if key is None:
        key_ = itemgetter(0)
    else:
        key_ = composition(itemgetter(0), key)
    ordered = sorted(zip(iterable, count(0)), key=key_, reverse=reverse)
    return list(map(itemgetter(1), ordered))


def linspace(count):
    # type: (int) -> Iterable[float]
    """
    Return `count` evenly spaced points from 0..1 interval.

    >>> list(linspace(3)))
    [0.0, 0.5, 1.0]
    """
    if count > 1:
        return (i / (count - 1) for i in range(count))
    elif count == 1:
        return (_ for _ in (0.0,))
    elif count == 0:
        return (_ for _ in ())
    else:
        raise ValueError("Count must be non-negative")


def linspace_trunc(count):
    # type: (int) -> Iterable[float]
    """
    Return `count` evenly spaced points from 0..1 interval *excluding*
    both end points.

    >>> list(linspace_trunc(3))
    [0.25, 0.5, 0.75]
    """
    return islice(linspace(count + 2), 1, count + 1)


def sample_path(path, num=10):
    # type: (QPainterPath, int) -> List[QPointF]
    """
    Sample `num` equidistant points from the `path` (`QPainterPath`).
    """
    return [path.pointAtPercent(p) for p in linspace(num)]


def clip(a, amin, amax):
    """Clip (limit) the value `a` between `amin` and `amax`"""
    return max(min(a, amax), amin)


def saturated(color, factor=150):
    # type: (QColor, int) -> QColor
    """Return a saturated color.
    """
    h = color.hsvHueF()
    s = color.hsvSaturationF()
    v = color.valueF()
    a = color.alphaF()
    s = factor * s / 100.0
    s = clip(s, 0.0, 1.0)
    return QColor.fromHsvF(h, s, v, a).convertTo(color.spec())


def radial_gradient(color, color_light=50):
    # type: (QColor, Union[int, QColor]) -> QRadialGradient
    """
    radial_gradient(QColor, QColor)
    radial_gradient(QColor, int)

    Return a radial gradient. `color_light` can be a QColor or an int.
    In the later case the light color is derived from `color` using
    `saturated(color, color_light)`.

    """
    if not isinstance(color_light, QColor):
        color_light = saturated(color, color_light)
    gradient = QRadialGradient(0.5, 0.5, 0.5)
    gradient.setColorAt(0.0, color_light)
    gradient.setColorAt(0.5, color_light)
    gradient.setColorAt(1.0, color)
    gradient.setCoordinateMode(QRadialGradient.ObjectBoundingMode)
    return gradient


def toGraphicsObjectIfPossible(item):
    """Return the item as a QGraphicsObject if possible.

    This function is intended as a workaround for a problem with older
    versions of PyQt (< 4.9), where methods returning 'QGraphicsItem *'
    lose the type of the QGraphicsObject subclasses and instead return
    generic QGraphicsItem wrappers.
    """
    if item is None:
        return None

    obj = item.toGraphicsObject()
    return item if obj is None else obj


def uniform_linear_layout_trunc(points):
    # type: (List[float]) -> List[float]
    """
    Layout the points (a list of floats in 0..1 range) in a uniform
    linear space (truncated) while preserving the existing sorting order.
    """
    indices = argsort(points)
    indices = invert_permutation_indices(indices)
    space = list(linspace_trunc(len(indices)))
    return [space[i] for i in indices]


def invert_permutation_indices(indices):
    # type: (List[int]) -> List[int]
    """
    Invert the permutation given by indices.
    """
    inverted = [sys.maxsize] * len(indices)
    for i, index in enumerate(indices):
        inverted[index] = i
    return inverted


def stroke_path(path, pen):
    # type: (QPainterPath, QPen) -> QPainterPath
    """Create a QPainterPath stroke from the `path` drawn with `pen`.
    """
    stroker = QPainterPathStroker()
    stroker.setCapStyle(pen.capStyle())
    stroker.setJoinStyle(pen.joinStyle())
    stroker.setMiterLimit(pen.miterLimit())
    stroker.setWidth(max(pen.widthF(), 1e-9))

    return stroker.createStroke(path)


def bezier_subdivide(cp, t):
    # type: (List[QPointF], float) -> Tuple[List[QPointF], List[QPointF]]
    """
    Subdivide a cubic bezier curve defined by the control points `cp`.

    Parameters
    ----------
    cp : List[QPointF]
        The control points for a cubic bezier curve.
    t : float
        The cut point; a value between 0 and 1.

    Returns
    -------
    cp : Tuple[List[QPointF], List[QPointF]]
        Two lists of new control points for the new left and right part
        respectively.
    """
    # http://www.cs.mtu.edu/~shene/COURSES/cs3621/NOTES/spline/Bezier/bezier-sub.html
    c00, c01, c02, c03 = cp

    c10 = c00 * (1 - t) + c01 * t
    c11 = c01 * (1 - t) + c02 * t
    c12 = c02 * (1 - t) + c03 * t

    c20 = c10 * (1 - t) + c11 * t
    c21 = c11 * (1 - t) + c12 * t

    c30 = c20 * (1 - t) + c21 * t

    first = [c00, c10, c20, c30]
    second = [c30, c21, c12, c03]
    return first, second


class _Section(typing.NamedTuple):
    #: Single section path (single cubicTo, lineTo, moveTo path)
    p: QPainterPath
    #: The path section type
    type: QPainterPath.ElementType
    #: The approximate original start distance
    start: float
    #: The approximate original end distance
    end: float


def _qpainterpath_sections(path: QPainterPath) -> Iterable[_Section]:
    """
    Return `path` elementary *sections* (single line, single bezier or
    move elements).
    """
    if path.isEmpty():
        return
    el0 = path.elementAt(0)
    assert el0.type == QPainterPath.MoveToElement
    i = 1
    section_start = section_end = 0
    while i < path.elementCount():
        el1 = path.elementAt(i)
        if el1.type == QPainterPath.LineToElement:
            p = QPainterPath()
            p.moveTo(el0.x, el0.y)
            p.lineTo(el1.x, el1.y)
            section_end += p.length()
            yield _Section(p, QPainterPath.LineToElement, section_start, section_end)
            i += 1
            el0 = el1
            section_start = section_end
        elif el1.type == QPainterPath.CurveToElement:
            c0, c1, c2, c3 = el0, el1, path.elementAt(i + 1), path.elementAt(i + 2)
            assert all(el.type == QPainterPath.CurveToDataElement
                       for el in [c2, c3])
            p0, p1, p2, p3 = [QPointF(el.x, el.y) for el in [c0, c1, c2, c3]]

            p = QPainterPath()
            p.moveTo(p0)
            p.cubicTo(p1, p2, p3)
            section_end += p.length()
            yield _Section(p, QPainterPath.CurveToElement, section_start, section_end)
            i += 3
            el0 = c3
            section_start = section_end
        elif el1.type == QPainterPath.MoveToElement:
            p = QPainterPath()
            p.moveTo(el1.x, el1.y)
            i += 1
            el0 = el1
            yield _Section(p, QPainterPath.MoveToElement, section_start, section_end)
            section_start = section_end


def _qpainterpath_simple_cut(path: QPainterPath, start: float, end: float):
    """
    Cut a sub path from a simple `path` (single lineTo or cubicTo).
    """
    assert 0. <= start <= end <= 1.0
    if path.elementCount() == 0:
        return QPainterPath()
    el0 = path.elementAt(0)
    assert el0.type == QPainterPath.MoveToElement
    if path.elementCount() == 1:
        return QPainterPath(path)
    el1 = path.elementAt(1)
    if el1.type == QPainterPath.LineToElement:
        segment = QLineF(el0.x, el0.y, el1.x, el1.y)
        p1 = segment.pointAt(start)
        p2 = segment.pointAt(end)
        p = QPainterPath()
        p.moveTo(p1)
        p.lineTo(p2)
        return p
    elif el1.type == QPainterPath.CurveToElement:
        c0, c1, c2, c3 = el0, el1, path.elementAt(2), path.elementAt(3)
        assert all(el.type == QPainterPath.CurveToDataElement
                   for el in [c2, c3])
        cp = [QPointF(el.x, el.y) for el in [c0, c1, c2, c3]]
        # adjust the end
        # |---------+---------+-----|
        # |--------start-----end----|
        end_ = (end - start) / (1 - start)
        assert 0 <= end_ <= 1.0
        _, cp = bezier_subdivide(cp, start)
        cp, _ = bezier_subdivide(cp, end_)
        p = QPainterPath()
        p.moveTo(cp[0])
        p.cubicTo(*cp[1:])
        return p
    else:
        assert False


def qpainterpath_sub_path(
        path: QPainterPath, start: float, end: float
) -> QPainterPath:
    """
    Cut and return a sub path from `path`.

    Parameters
    ----------
    path: QPainterPath
        The source path.
    start: float
        The starting position for the cut as a number between `0.0` and `1.0`
    end: float
        The end position for the cut as a number between `0.0` and `1.0`
    """
    assert 0.0 <= start <= 1.0 and 0.0 <= end <= 1.0
    length = path.length()
    startlen = length * start
    endlen = length * end
    res = QPainterPath()
    for section in list(_qpainterpath_sections(path)):
        if startlen <= section.start <= endlen \
                or startlen <= section.end <= endlen \
                or (section.start <= startlen and section.end >= endlen):
            if math.isclose(section.p.length(), 0):
                res.addPath(section.p)
            else:
                start_ = (startlen - section.start) / section.p.length()
                end_ = (endlen - section.start) / section.p.length()
                p = _qpainterpath_simple_cut(
                    section.p, max(start_, 0.), min(end_, 1.0)
                )
                res.addPath(p)
    return res



from typing import Sequence

from AnyQt.QtCore import QObject, Signal, Slot, Qt
from AnyQt.QtWidgets import QWidget, QAction, QActionGroup, QApplication

from orangecanvas.utils import findf

__all__ = [
    "WindowListManager",
]


class WindowListManager(QObject):
    """
    An open windows list manager.

    Provides and manages actions for opened 'Windows' menu bar entries.
    """
    #: Signal emitted when a widget/window is added
    windowAdded = Signal(QWidget, QAction)
    #: Signal emitted when a widget/window is removed
    windowRemoved = Signal(QWidget, QAction)

    __instance = None

    @staticmethod
    def instance() -> "WindowListManager":
        """Return the global WindowListManager instance."""
        if WindowListManager.__instance is None:
            return WindowListManager()
        return WindowListManager.__instance

    def __init__(self, *args, **kwargs):
        if self.__instance is not None:
            raise RuntimeError
        WindowListManager.__instance = self
        super().__init__(*args, **kwargs)
        self.__group = QActionGroup(
            self, objectName="window-list-manager-action-group"
        )
        self.__group.setExclusive(True)
        self.__windows = []
        app = QApplication.instance()
        app.focusWindowChanged.connect(
            self.__focusWindowChanged, Qt.QueuedConnection
        )

    def actionGroup(self) -> QActionGroup:
        """Return the QActionGroup containing the *Window* actions."""
        return self.__group

    def addWindow(self, window: QWidget) -> None:
        """Add a `window` to the managed list."""
        if window in self.__windows:
            raise ValueError(f"{window} already added")
        action = self.createActionForWindow(window)
        self.__windows.append(window)
        self.__group.addAction(action)
        self.windowAdded.emit(window, action)

    def removeWindow(self, window: QWidget) -> None:
        """Remove the `window` from the managed list."""
        self.__windows.remove(window)
        act = self.actionForWindow(window)
        self.__group.removeAction(act)
        self.windowRemoved.emit(window, act)
        act.setData(None)
        act.setParent(None)
        act.deleteLater()

    def actionForWindow(self, window: QWidget) -> QAction:
        """Return the `QAction` representing the `window`."""
        return findf(self.actions(), lambda a: a.data() is window)

    def createActionForWindow(self, window: QWidget) -> QAction:
        """Create the `QAction` instance for managing the `window`."""
        action = QAction(
            window.windowTitle(),
            window,
            visible=window.isVisible(),
            checkable=True,
            objectName="action-canvas-window-list-manager-window-action"
        )
        action.setData(window)
        handle = window.windowHandle()
        if not handle:
            # TODO: need better visible, title notify bypassing QWindow
            window.create()
            handle = window.windowHandle()
        action.setChecked(handle.isActive())
        handle.visibleChanged.connect(action.setVisible)
        handle.windowTitleChanged.connect(action.setText)

        def activate(state):
            if not state:
                return
            handle: QWidget = action.data()
            handle.setVisible(True)
            if handle != QApplication.activeWindow():
                # Do not re-activate when called from `focusWindowChanged`;
                # breaks macOS window cycling (CMD+`) order.
                handle.raise_()
                handle.activateWindow()

        action.toggled.connect(activate)
        return action

    def actions(self) -> Sequence[QAction]:
        """Return all actions representing managed windows."""
        return self.__group.actions()

    @Slot()
    def __focusWindowChanged(self):
        window = QApplication.activeWindow()
        act = findf(self.actions(), lambda a: a.data() is window)
        if act is not None and not act.isChecked():
            act.setChecked(True)



import itertools
import json
import logging
import os
import re
import shlex
import sys
import sysconfig
from collections import deque
from enum import Enum
from types import SimpleNamespace
from typing import AnyStr, Callable, List, NamedTuple, Optional, Tuple, TypeVar, Union

import requests
from AnyQt.QtCore import QObject, QSettings, QStandardPaths, QTimer, Signal, Slot
from pkg_resources import (
    Requirement,
    ResolutionError,
    VersionConflict,
    WorkingSet,
    get_distribution,
    parse_version,
)

from orangecanvas.utils import unique
from orangecanvas.utils.pkgmeta import parse_meta
from orangecanvas.utils.shtools import create_process, python_process

log = logging.getLogger(__name__)

PYPI_API_JSON = "https://pypi.org/pypi/{name}/json"
A = TypeVar("A")
B = TypeVar("B")


def normalize_name(name):
    return re.sub(r"[-_.]+", "-", name).lower()


def prettify_name(name):
    dash_split = name.split('-')
    # Orange3-ImageAnalytics => ImageAnalytics
    orange_prefix = len(dash_split) > 1 and dash_split[0].lower() in ['orange', 'orange3']
    name = ' '.join(dash_split[1:] if orange_prefix else dash_split)
    # ImageAnalytics => Image Analytics  # while keeping acronyms
    return re.sub(r"(?<!^)((?<![\s\d])[A-Z][a-z]|(?<=[a-z])[A-Z])", r" \1", name)


class Installable(
    NamedTuple(
        "Installable", (
            ("name", str),
            ("version", str),
            ("summary", str),
            ("description", str),
            ("package_url", str),
            ("release_urls", List['ReleaseUrl']),
            ("requirements", List[str]),
            ("description_content_type", Optional[str]),
            ("force", bool),
        ))):
    """
    An installable distribution from PyPi

    Attributes
    ----------
    name: str
        The distribution/project name
    version: str
        The release version
    summary: str
        Short one line summary text
    description: str
        A longer more detailed description
    package_url: str
    release_urls: List[ReleaseUrls]
    """

Installable.__new__.__defaults__ = (
    [],
    None,  # description_content_type = None,
    False,  # force = False
)


class ReleaseUrl(
    NamedTuple(
        "ReleaseUrl", (
            ("filename", str),
            ("url", str),
            ("size", int),
            ("python_version", str),
            ("package_type", str),
        ))):
    """
    An source/wheel/egg release for a distribution,
    """


class Available(
    NamedTuple(
        "Available", (
            ("installable", Installable),
    ))):
    """
    An available package.

    Attributes
    ----------
    installable : Installable
    """
    @property
    def project_name(self):
        return self.installable.name

    @property
    def normalized_name(self):
        return normalize_name(self.project_name)


class Installed(
    NamedTuple(
        "Installed", (
            ("installable", Optional[Installable]),
            ("local", 'Distribution'),
            ("required", bool),
            ("constraint", Optional[Requirement]),
        ))):
    """
    An installed package. Does not need to have a corresponding installable
    entry (eg. only local or private distribution)

    Attributes
    ----------
    installable: Installable
        An optional installable item. Is None if the package is not available
        from any package index (is not published and installed locally or
        possibly orphaned).
    local : Distribution
        A :class:`~.Distribution` instance representing the distribution meta
        of the locally installed package.
    required : bool
        Is the distribution required (is part of the core application and
        must not be uninstalled).
    constraint: Optional[Requirement]
        A version constraint string.
    """
    def __new__(cls, installable, local, required=False, constraint=None):
        # type: (Optional[Installable], Distribution, bool, Optional[Requirement]) -> Installed
        return super().__new__(cls, installable, local, required, constraint)

    @property
    def project_name(self):
        if self.installable is not None:
            return self.installable.name
        else:
            return self.local.project_name

    @property
    def normalized_name(self):
        return normalize_name(self.project_name)


#: An installable item/slot
Item = Union[Available, Installed]


def is_updatable(item):
    # type: (Item) -> bool
    if isinstance(item, Available):
        return False
    elif item.installable is None:
        return False
    else:
        inst, dist = item.installable, item.local
        try:
            v1 = parse_version(dist.version)
            v2 = parse_version(inst.version)
        except ValueError:
            return False

        if inst.force:
            return True

        if item.constraint is not None and str(v2) not in item.constraint:
            return False
        else:
            return v1 < v2


def get_meta_from_archive(path):
    """Return project metadata extracted from sdist or wheel archive, or None
    if metadata can't be found."""

    def is_metadata(fname):
        return fname.endswith(('PKG-INFO', 'METADATA'))

    meta = None
    if path.endswith(('.zip', '.whl')):
        from zipfile import ZipFile
        with ZipFile(path) as archive:
            meta = next(filter(is_metadata, archive.namelist()), None)
            if meta:
                meta = archive.read(meta).decode('utf-8')
    elif path.endswith(('.tar.gz', '.tgz')):
        import tarfile
        with tarfile.open(path) as archive:
            meta = next(filter(is_metadata, archive.getnames()), None)
            if meta:
                meta = archive.extractfile(meta).read().decode('utf-8')
    if meta:
        return parse_meta(meta)


def pypi_json_query_project_meta(projects, session=None):
    # type: (List[str], Optional[requests.Session]) -> List[Optional[dict]]
    """
    Parameters
    ----------
    projects : List[str]
        List of project names to query
    session : Optional[requests.Session]
    """
    if session is None:
        session = _session()
    rval = []  # type: List[Optional[dict]]
    for name in projects:
        r = session.get(PYPI_API_JSON.format(name=name))
        if r.status_code != 200:
            rval.append(None)
        else:
            try:
                meta = r.json()
            except json.JSONDecodeError:
                rval.append(None)
            else:
                try:
                    # sanity check
                    installable_from_json_response(meta)
                except (TypeError, KeyError):
                    rval.append(None)
                else:
                    rval.append(meta)
    return rval


def installable_from_json_response(meta):
    # type: (dict) -> Installable
    """
    Extract relevant project meta data from a PyPiJSONRPC response

    Parameters
    ----------
    meta : dict
        JSON response decoded into python native dict.

    Returns
    -------
    installable : Installable
    """
    info = meta["info"]
    name = info["name"]
    version = info.get("version", "0")
    summary = info.get("summary", "")
    description = info.get("description", "")
    content_type = info.get("description_content_type", None)
    package_url = info.get("package_url", "")
    distributions = meta.get("releases", {}).get(version, [])
    release_urls = [ReleaseUrl(r["filename"], url=r["url"], size=r["size"],
                               python_version=r.get("python_version", ""),
                               package_type=r["packagetype"])
                    for r in distributions]
    requirements = info.get("requires_dist", [])

    return Installable(name, version, summary, description, package_url, release_urls,
                       requirements, content_type)


def _session(cachedir=None):
    # type: (...) -> requests.Session
    """
    Return a requests.Session instance

    Parameters
    ----------
    cachedir : Optional[str]
        HTTP cache location.

    Returns
    -------
    session : requests.Session
    """
    import cachecontrol.caches
    if cachedir is None:
        cachedir = QStandardPaths.writableLocation(QStandardPaths.CacheLocation)
        cachedir = os.path.join(cachedir, "networkcache", "requests")
    session = requests.Session()
    session = cachecontrol.CacheControl(
        session, cache=cachecontrol.caches.FileCache(directory=cachedir)
    )
    return session


def optional_map(func: Callable[[A], B]) -> Callable[[Optional[A]], Optional[B]]:
    def f(x: Optional[A]) -> Optional[B]:
        return func(x) if x is not None else None
    return f


class _QueryResult(SimpleNamespace):
    def __init__(
            self, queryname: str, installable: Optional[Installable], **kwargs
    ) -> None:
        self.queryname = queryname
        self.installable = installable
        super().__init__(**kwargs)


def query_pypi(names: List[str]) -> List[_QueryResult]:
    res = pypi_json_query_project_meta(names)
    installable_from_json_response_ = optional_map(
        installable_from_json_response
    )
    return [
        _QueryResult(name, installable_from_json_response_(r))
        for name, r in zip(names, res)
    ]


def list_available_versions(config, session=None):
    # type: (config.Config, Optional[requests.Session]) -> (List[Installable], List[Exception])
    if session is None:
        session = _session()

    exceptions = []

    try:
        defaults = config.addon_defaults_list()
    except requests.exceptions.RequestException as e:
        defaults = []
        exceptions.append(e)

    def getname(item):
        # type: (Dict[str, Any]) -> str
        info = item.get("info", {})
        if not isinstance(info, dict):
            return ""
        name = info.get("name", "")
        assert isinstance(name, str)
        return name

    defaults_names = {getname(a) for a in defaults}

    # query pypi.org for installed add-ons that are not in the defaults
    # list
    installed = [ep.dist for ep in config.addon_entry_points()
                 if ep.dist is not None]
    missing = {dist.project_name.casefold() for dist in installed} - \
              {name.casefold() for name in defaults_names}

    distributions = []
    for p in missing:
        try:
            response = session.get(PYPI_API_JSON.format(name=p))
            if response.status_code != 200:
                continue
            distributions.append(response.json())
        except requests.exceptions.RequestException as e:
            exceptions.append(e)

    packages = []
    for addon in distributions + defaults:
        try:
            packages.append(installable_from_json_response(addon))
        except (TypeError, KeyError) as e:
            exceptions.append(e)

    return packages, exceptions


def installable_items(pypipackages, installed=[]):
    # type: (Iterable[Installable], Iterable[Distribution]) -> List[Item]
    """
    Return a list of installable items.

    Parameters
    ----------
    pypipackages : list of Installable
    installed : list of pkg_resources.Distribution
    """

    dists = {dist.project_name: dist for dist in installed}
    packages = {pkg.name: pkg for pkg in pypipackages}

    # For every pypi available distribution not listed by
    # `installed`, check if it is actually already installed.
    ws = WorkingSet()
    for pkg_name in set(packages.keys()).difference(set(dists.keys())):
        try:
            d = ws.find(Requirement.parse(pkg_name))
        except ResolutionError:
            pass
        except ValueError:
            # Requirements.parse error ?
            pass
        else:
            if d is not None:
                dists[d.project_name] = d

    project_names = unique(itertools.chain(packages.keys(), dists.keys()))

    items = []  # type: List[Item]
    for name in project_names:
        if name in dists and name in packages:
            item = Installed(packages[name], dists[name])
        elif name in dists:
            item = Installed(None, dists[name])
        elif name in packages:
            item = Available(packages[name])
        else:
            assert False
        items.append(item)
    return items


def is_requirement_available(
        req: Union[Requirement, str], working_set: Optional[WorkingSet] = None
) -> bool:
    if not isinstance(req, Requirement):
        req = Requirement.parse(req)
    try:
        if working_set is None:
            d = get_distribution(req)
        else:
            d = working_set.find(req)
    except VersionConflict:
        return False
    except ResolutionError:
        return False
    else:
        return d is not None


def have_install_permissions():
    """Check if we can create a file in the site-packages folder.
    This works on a Win7 miniconda install, where os.access did not. """
    try:
        fn = os.path.join(sysconfig.get_path("purelib"), "test_write_" + str(os.getpid()))
        with open(fn, "w"):
            pass
        os.remove(fn)
        return True
    except PermissionError:
        return False
    except OSError:
        return False


class Command(Enum):
    Install = "Install"
    Upgrade = "Upgrade"
    Uninstall = "Uninstall"


Install = Command.Install
Upgrade = Command.Upgrade
Uninstall = Command.Uninstall

Action = Tuple[Command, Item]


class CommandFailed(Exception):
    def __init__(self, cmd, retcode, output):
        if not isinstance(cmd, str):
            cmd = " ".join(map(shlex.quote, cmd))
        self.cmd = cmd
        self.retcode = retcode
        self.output = output


class Installer(QObject):
    installStatusChanged = Signal(str)
    started = Signal()
    finished = Signal()
    error = Signal(str, object, int, list)

    def __init__(self, parent=None, steps=[]):
        super().__init__(parent)
        self.__interupt = False
        self.__queue = deque(steps)
        self.__statusMessage = ""
        self.pip = PipInstaller()
        self.conda = CondaInstaller()

    def start(self):
        QTimer.singleShot(0, self._next)

    def interupt(self):
        self.__interupt = True

    def setStatusMessage(self, message):
        if self.__statusMessage != message:
            self.__statusMessage = message
            self.installStatusChanged.emit(message)

    @Slot()
    def _next(self):
        command, pkg = self.__queue.popleft()
        try:
            if command == Install \
                    or (command == Upgrade and pkg.installable.force):
                self.setStatusMessage(
                    "Installing {}".format(pkg.installable.name))
                if self.conda:
                    try:
                        self.conda.install(pkg.installable)
                    except CommandFailed:
                        self.pip.install(pkg.installable)
                else:
                    self.pip.install(pkg.installable)
            elif command == Upgrade:
                self.setStatusMessage(
                    "Upgrading {}".format(pkg.installable.name))
                if self.conda:
                    try:
                        self.conda.upgrade(pkg.installable)
                    except CommandFailed:
                        self.pip.upgrade(pkg.installable)
                else:
                    self.pip.upgrade(pkg.installable)
            elif command == Uninstall:
                self.setStatusMessage(
                    "Uninstalling {}".format(pkg.local.project_name))
                if self.conda:
                    try:
                        self.conda.uninstall(pkg.local)
                    except CommandFailed:
                        self.pip.uninstall(pkg.local)
                else:
                    self.pip.uninstall(pkg.local)
        except CommandFailed as ex:
            self.error.emit(
                "Command failed: python {}".format(ex.cmd),
                pkg, ex.retcode, ex.output
            )
            return

        if self.__queue:
            QTimer.singleShot(0, self._next)
        else:
            self.finished.emit()


class PipInstaller:
    def __init__(self):
        arguments = QSettings().value('add-ons/pip-install-arguments', '', type=str)
        self.arguments = shlex.split(arguments)

    def install(self, pkg):
        # type: (Installable) -> None
        cmd = [
            "python", "-m", "pip", "install", "--upgrade",
            "--upgrade-strategy=only-if-needed",
        ] + self.arguments
        if pkg.package_url.startswith(("http://", "https://")):
            version = "=={}".format(pkg.version) if pkg.version is not None else ""
            cmd.append(pkg.name + version)
        else:
            # Package url is path to the (local) wheel
            cmd.append(pkg.package_url)
        run_command(cmd)

    def upgrade(self, package):
        cmd = [
            "python", "-m", "pip", "install",
                "--upgrade", "--upgrade-strategy=only-if-needed",
        ] + self.arguments
        if package.package_url.startswith(("http://", "https://")):
            version = (
                "=={}".format(package.version) if package.version is not None
                else ""
            )
            cmd.append(package.name + version)
        else:
            cmd.append(package.package_url)
        run_command(cmd)

    def uninstall(self, dist):
        cmd = ["python", "-m", "pip", "uninstall", "--yes", dist.project_name]
        run_command(cmd)


class CondaInstaller:
    def __init__(self):
        enabled = QSettings().value('add-ons/allow-conda', True, type=bool)
        if enabled:
            self.conda = self._find_conda()
        else:
            self.conda = None

    def _find_conda(self):
        executable = sys.executable
        bin = os.path.dirname(executable)

        # posix
        conda = os.path.join(bin, "conda")
        if os.path.exists(conda):
            return conda

        # windows
        conda = os.path.join(bin, "Scripts", "conda.bat")
        if os.path.exists(conda):
            # "activate" conda environment orange is running in
            os.environ["CONDA_PREFIX"] = bin
            os.environ["CONDA_DEFAULT_ENV"] = bin
            return conda

    def install(self, pkg):
        version = "={}".format(pkg.version) if pkg.version is not None else ""
        cmd = [self.conda, "install", "--yes", "--quiet",
               "--satisfied-skip-solve",
               self._normalize(pkg.name) + version]
        return run_command(cmd)

    def upgrade(self, pkg):
        version = "={}".format(pkg.version) if pkg.version is not None else ""
        cmd = [self.conda, "install", "--yes", "--quiet",
               "--satisfied-skip-solve",
               self._normalize(pkg.name) + version]
        return run_command(cmd)

    def uninstall(self, dist):
        cmd = [self.conda, "uninstall", "--yes",
               self._normalize(dist.project_name)]
        return run_command(cmd)

    def _normalize(self, name):
        # Conda 4.3.30 is inconsistent, upgrade command is case sensitive
        # while install and uninstall are not. We assume that all conda
        # package names are lowercase which fixes the problems (for now)
        return name.lower()

    def __bool__(self):
        return bool(self.conda)


def run_command(command, raise_on_fail=True, **kwargs):
    # type: (List[str], bool, Any) -> Tuple[int, List[AnyStr]]
    """
    Run command in a subprocess.

    Return `process` return code and output once it completes.
    """
    log.info("Running %s", " ".join(command))

    if command[0] == "python":
        process = python_process(command[1:], **kwargs)
    else:
        process = create_process(command, **kwargs)
    rcode, output = run_process(process, file=sys.stdout)
    if rcode != 0 and raise_on_fail:
        raise CommandFailed(command, rcode, output)
    else:
        return rcode, output


def run_process(process: 'subprocess.Popen', **kwargs) -> Tuple[int, List[AnyStr]]:
    file = kwargs.pop("file", sys.stdout)  # type: Optional[IO]
    if file is ...:
        file = sys.stdout

    output = []
    while process.poll() is None:
        line = process.stdout.readline()
        output.append(line)
        print(line, end="", file=file)
    # Read remaining output if any
    line = process.stdout.read()
    if line:
        output.append(line)
        print(line, end="", file=file)
    return process.returncode, output



"""
A LineEdit class with a button on left/right side.
"""

from collections import namedtuple

from typing import Any, Optional, List, NamedTuple

from AnyQt.QtWidgets import (
    QLineEdit, QToolButton, QStyleOptionToolButton, QStylePainter,
    QStyle, QAction, QWidget,
)
from AnyQt.QtGui import QPaintEvent, QPainter, QColor
from AnyQt.QtCore import Qt, QSize, QRect
from AnyQt.QtCore import pyqtSignal as Signal, pyqtProperty as Property

from orangecanvas.gui.utils import innerShadowPixmap

_ActionSlot = NamedTuple(
    "_ActionSlot", [
        ("position", 'int'),  # Left/Right position
        ("action", 'QAction'),  # QAction
        ("button", 'LineEditButton'),   # LineEditButton instance
        ("autoHide", 'Any'),  # Auto hide when line edit is empty (unused??)
    ]
)


class LineEditButton(QToolButton):
    """
    A button in the :class:`LineEdit`.
    """
    def __init__(self, parent=None, flat=True, **kwargs):
        # type: (Optional[QWidget], bool, Any) -> None
        super().__init__(parent, **kwargs)
        self.__flat = flat
        self.__shadowLength = 5
        self.__shadowPosition = 0
        self.__shadowColor = QColor("#000000")

    def setFlat(self, flat):
        # type: (bool) -> None
        if self.__flat != flat:
            self.__flat = flat
            self.update()

    def flat(self):
        # type: () -> bool
        return self.__flat

    flat_ = Property(bool, fget=flat, fset=setFlat,
                     designable=True)

    def setShadowLength(self, shadowSize):
        if self.__shadowLength != shadowSize:
            self.__shadowLength = shadowSize
            self.update()

    def shadowLength(self):
        return self.__shadowLength

    shadowLength_ = Property(int, fget=shadowLength, fset=setShadowLength, designable=True)

    def setShadowPosition(self, shadowPosition):
        if self.__shadowPosition != shadowPosition:
            self.__shadowPosition = shadowPosition
            self.update()

    def shadowPosition(self):
        return self.__shadowPosition

    shadowPosition_ = Property(int, fget=shadowPosition, fset=setShadowPosition, designable=True)

    def setShadowColor(self, shadowColor):
        if self.__shadowColor != shadowColor:
            self.__shadowColor = shadowColor
            self.update()

    def shadowColor(self):
        return self.__shadowColor

    shadowColor_ = Property(QColor, fget=shadowColor, fset=setShadowColor, designable=True)

    def paintEvent(self, event):
        # type: (QPaintEvent) -> None
        if self.__flat:
            opt = QStyleOptionToolButton()
            self.initStyleOption(opt)
            p = QStylePainter(self)
            p.drawControl(QStyle.CE_ToolButtonLabel, opt)
            p.end()
        else:
            super().paintEvent(event)

        # paint shadow
        shadow = innerShadowPixmap(self.__shadowColor,
                                   self.size(),
                                   self.__shadowPosition,
                                   length=self.__shadowLength)

        p = QPainter(self)

        rect = self.rect()
        targetRect = QRect(rect.left() + 1,
                           rect.top() + 1,
                           rect.width() - 2,
                           rect.height() - 2)

        p.drawPixmap(targetRect, shadow, shadow.rect())
        p.end()


class LineEdit(QLineEdit):
    """
    A line edit widget with support for adding actions (buttons) to
    the left/right of the edited text

    """
    #: Position flags
    LeftPosition, RightPosition = 1, 2

    #: Emitted when the action is triggered.
    triggered = Signal(QAction)

    #: The left action was triggered.
    leftTriggered = Signal()

    #: The right action was triggered.
    rightTriggered = Signal()

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None
        super().__init__(*args, **kwargs)
        self.__actions = [None, None]  # type: List[Optional[_ActionSlot]]

    def setAction(self, action, position=LeftPosition):
        # type: (QAction, int) -> None
        """
        Set `action` to be displayed at `position`. Existing action
        (if present) will be removed.

        Parameters
        ----------
        action : :class:`QAction`
        position : int
            Position where to set the action (default: ``LeftPosition``).
        """
        curr = self.actionAt(position)
        if curr is not None:
            self.removeActionAt(position)

        # Add the action using QWidget.addAction (for shortcuts)
        self.addAction(action)

        button = LineEditButton(self)
        button.setToolButtonStyle(Qt.ToolButtonIconOnly)
        button.setDefaultAction(action)
        button.setVisible(self.isVisible())
        button.show()
        button.setCursor(Qt.ArrowCursor)

        button.triggered.connect(self.triggered)
        button.triggered.connect(self.__onTriggered)

        slot = _ActionSlot(position, action, button, False)
        self.__actions[position - 1] = slot

        if not self.testAttribute(Qt.WA_Resized):
            # Need some sensible height to do the layout.
            self.adjustSize()

        self.__layoutActions()

    def actionAt(self, position):
        # type: (int) -> Optional[QAction]
        """
        Return :class:`QAction` at `position`.
        """
        self._checkPosition(position)
        slot = self.__actions[position - 1]
        if slot:
            return slot.action
        else:
            return None

    def removeActionAt(self, position):
        # type: (int) -> None
        """
        Remove the action at position.
        """
        self._checkPosition(position)

        slot = self.__actions[position - 1]
        self.__actions[position - 1] = None
        if slot is not None:
            slot.button.hide()
            slot.button.deleteLater()
            self.removeAction(slot.action)
            self.__layoutActions()

    def button(self, position):
        # type: (int) -> Optional[LineEditButton]
        """
        Return the button (:class:`LineEditButton`) for the action
        at `position`.

        """
        self._checkPosition(position)
        slot = self.__actions[position - 1]
        if slot is not None:
            return slot.button
        else:
            return None

    def _checkPosition(self, position):
        # type: (int) -> None
        if position not in [self.LeftPosition, self.RightPosition]:
            raise ValueError("Invalid position")

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.__layoutActions()

    def __layoutActions(self):  # type: () -> None
        left, right = self.__actions

        contents = self.contentsRect()
        buttonSize = QSize(contents.height(), contents.height())

        margins = self.textMargins()

        if left:
            geom = QRect(contents.topLeft(), buttonSize)
            left.button.setGeometry(geom)
            margins.setLeft(buttonSize.width())

        if right:
            geom = QRect(contents.topRight(), buttonSize)
            right.button.setGeometry(geom.translated(-buttonSize.width(), 0))
            margins.setLeft(buttonSize.width())

        self.setTextMargins(margins)

    def __onTriggered(self, action):
        # type: (QAction) -> None
        left, right = self.__actions
        if left and action == left.action:
            self.leftTriggered.emit()
        elif right and action == right.action:
            self.rightTriggered.emit()
